package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils._

/** Runge function and its derivative interpolation:
  * Least Squares and Radon-Nikodym. 
  * @version $Id: ExampleRungeFunction.scala,v 1.5 2015/11/04 09:58:59 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
  * A library from http://www.ioffe.ru/LNEPS/malyshkin/code_polynomials_quadratures.zip is required.
  */
class ExampleRungeFunction(val dim:Int) {
  val Q=new OrthogonalPolynomialsLegendreBasis()

  /** Runge function: 1/(1+25x^2)
    */
  val Runge=new PolynomialsRatio(Array(1,0,25),Array(1),new OrthogonalPolynomialsMonomialsBasis())
  val XMIN= -1.0
  val XMAX= 1.0
  val EPS=1e-5

  /** Calculate <f*Q> moments using numerical integration.
    */
  def getMoments(n:Int,fun: Double => Double)={
    val m=new Array[Double](n)
    val intg=new QUANC8.DoubleFunction{
      def f(x:Double):Double=fun(x)*Q.fx(m,x)
    }
    val mom=new Array[Double](n)
    for(k <-0 until n)
    {
      m(k)=1
      val i=QUANC8.integrate(intg,XMIN,XMAX,EPS,EPS,1000000);
      if(!(i.flag==0))
      {
	throw new RuntimeException("flag="+i.flag);
      }
      mom(k)=i.integral;
      m(k)=0
    }
    mom
  }

  def getConstMoments(n:Int)=getMoments(n:Int,(x) => 1)
  def getRungeMoments(n:Int)=getMoments(n:Int,(x) => Runge.getF(x))
  def getDRungeMoments(n:Int)=getMoments(n:Int,(x) => Runge.getDF(x))

  val Cmoms=getConstMoments(2*dim-1)
  val Fmoms=getRungeMoments(2*dim-1)
  val DFmoms=getDRungeMoments(2*dim-1)
  val IQQ=Linsystems.getInvertedMatrix(dim,dim,Q.getQQMatr(dim,Cmoms))

  val RUNGE_LS=TMatr.getMatr_x_A(dim,IQQ,dim,Fmoms)
  val DRUNGE_LS=TMatr.getMatr_x_A(dim,IQQ,dim,DFmoms)
  val RUNGE_RN=NevaiOperator.getNevaiOperator(dim,Cmoms,Fmoms,Q)
  val DRUNGE_RN=NevaiOperator.getNevaiOperator(dim,Cmoms,DFmoms,Q)
  
}

object ExampleRungeFunction {
  def main(args: Array[String]) {
    if(args.length!=1)
    {
      System.err.println("usage: ExampleRungeFunction dim")
      System.exit(1)
    }
    val dim=java.lang.Integer.valueOf(args(0))
    val E=new ExampleRungeFunction(dim)
    val dx=1e-4
    System.out.println("|x\tR\tLS_R\tRN_R\tDR\tLS_DR\tRN_DR\td/dx LS_R\td/dx RN_R")
    for(x <- E.XMIN to (E.XMAX+dx/2) by dx)
    {
      System.out.println(""+x+
        "\t"+E.Runge.getF(x)+"\t"+E.Q.fx(E.RUNGE_LS,x)+"\t"+E.RUNGE_RN.getF(x)+
        "\t"+E.Runge.getDF(x)+"\t"+E.Q.fx(E.DRUNGE_LS,x)+"\t"+E.DRUNGE_RN.getF(x)+
        "\t"+E.Q.Dfx(E.RUNGE_LS,x)+"\t"+E.RUNGE_RN.getDF(x))
    }
  }
}
