package com.polytechnik.itch;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */

/** A convertor of file to ITCH4 messages.
 * specification.
 *  @version $Id: FileFeeder.java,v 1.16 2011/10/15 15:06:11 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class FileFeeder {

    private static final boolean PRG_DEBUG=true;

    private final StreamProcessor parser;
    private final String fname;
    public FileFeeder(final StreamProcessor parser,
		      final String fname)
    {
	this.parser=parser;
	this.fname=fname;
    }

    private int processDataChunk(final byte [] buf,final int b_start,final int b_len)
    {
	if(b_len<SoupBinTCP.MESSAGE_PADDING)
	{
	    return 0; // empty string
	}
	// First two bytes are message length.
	final int m_length=MessageProcessor.getShortIntegerFromBytes(buf,b_start);
	if(m_length<=0)
	{
	    throw new IllegalStateException("m_length="+m_length);
	}
	final int lenp=m_length+SoupBinTCP.MESSAGE_PADDING;
	if(lenp>b_len)
	{
	    // to few data read, try next time
	    return 0;
	}
	// actually execute parser
	final int r=this.parser.process(buf,b_start,lenp);
	if(r!=lenp)
	{
	    throw new RuntimeException("Failed to process, broken invariant.");
	}
	return lenp;
    }


    public static java.io.InputStream getITCHStreamFromFile(final String file)
	throws java.io.IOException {
	final java.io.InputStream r;
	if(file.toLowerCase().endsWith(".zip"))
	{
	    // input is .zip archive. decompress it
	    final java.util.zip.ZipInputStream in=
		new java.util.zip.ZipInputStream(new java.io.FileInputStream(file));
	    in.getNextEntry();
	    r=new java.io.BufferedInputStream(in,SoupBinTCP.GZIP_BUFFER_SIZE);
	}
	else if(file.toLowerCase().endsWith(".gz"))
	{
	    // input is .gz archive. decompress it
	    r=new java.util.zip.GZIPInputStream(new java.io.FileInputStream(file),SoupBinTCP.GZIP_BUFFER_SIZE);
	}
	else
	{
	    r=new java.io.BufferedInputStream(new java.io.FileInputStream(file),SoupBinTCP.GZIP_BUFFER_SIZE);
	}
	return r;
    }


    public long processStream()	throws java.io.IOException {
	final long t0=System.currentTimeMillis();
	if(PRG_DEBUG)
        {
	    System.err.println("Reading data now="+new java.util.Date(t0));
	}

	final java.io.InputStream r=getITCHStreamFromFile(this.fname);

	long n=0,br=0;
	final byte [] buf=new byte [SoupBinTCP.INTERNAL_BUFFER_SIZE];
	final int thr_move=2+((buf.length*2)/3);

	gloop:for(int b_start=0,n_left=buf.length,b_len=0;;)
	{
	    if(b_start>=thr_move)
	    {
		// move to the beginning;
		System.arraycopy(buf,b_start,buf,0,b_len);
		n_left+=b_start;
		b_start=0;
	    }

	    if(n_left<=0)
	    {
		throw new RuntimeException("n_left="+n_left+" must be positive");
	    }

	    final int syms_read=r.read(buf,b_start+b_len,n_left);
	    if(syms_read<0)
	    {
		break;
	    }
	    n_left-=syms_read;
	    b_len+=syms_read;
	    if(b_start+b_len+n_left!=buf.length)
	    {
		throw new RuntimeException("Broken Invariant");
	    }

	    for(;;)
	    {
		final int syms_processed=processDataChunk(buf,b_start,b_len);
		/*
		System.err.println("syms_processed="+syms_processed+
				   " b_start="+b_start+
				   " b_len="+b_len+
				   " n_left="+n_left);
		*/
		if(syms_processed==0)
		{
		    // need to read more data
		    break;
		}
		else if(syms_processed<0)
		{
		    // parser sent "END" signal.
		    if(PRG_DEBUG)
		    {
			System.err.println("FileFeeder:: premature session end, parser returned "+
					   syms_processed+"; now="+new java.util.Date());
		    }
		    break gloop;
		}
		n++;
		br+=syms_processed;
		b_start+=syms_processed;
		b_len-=syms_processed;
		if(PRG_DEBUG)
		{
		    if(false && n%100000==0) // now print from process() instead.
		    {
			System.err.println("FileFeeder:: Read "+n+" records; "+
					   br+" bytes read; now="+new java.util.Date());
		    }
		}
	    }
	}
	r.close();
	if(PRG_DEBUG)
	{
	    final long t1=System.currentTimeMillis();
	    System.err.println("TOTAL "+n+" records read; bytes="+br+"; time diff="+
			       ((t1-t0)/1000L)+
			       " seconds. file="+this.fname+
			       " started="+new java.util.Date(t0)+
			       " now="+new java.util.Date(t1));
	}
	return n;
    }



    /** Unit test. */
    public static void main(String [] args) throws java.io.IOException
    {
	System.err.println("Processing stream by $Id: FileFeeder.java,v 1.16 2011/10/15 15:06:11 mal Exp $. now="+new java.util.Date());
	if(args.length!=1)
	{
	    System.err.println("usage: FileFeeder.main input_itch41.gz");
	    System.exit(1);
	}
	final String fin=args[0];
	final String fname="xx.out.gz"; // save to file xx.out.gz

	System.err.println("FileFeeder::ITCH data to be read from file="+fin+" and saved to file="+fname);
	final java.io.OutputStream out=
	    new java.util.zip.GZIPOutputStream(new java.io.FileOutputStream(fname),SoupBinTCP.GZIP_BUFFER_SIZE);

	new FileFeeder(new StreamProcessorDumpData(out,new ITCHParser(new DummyTrader())),
		       fin).processStream();

	out.close();
	System.err.println("FileFeeder::ITCH data read from file="+fin+" is saved to file="+fname);

	System.err.println("DONE processing stream by $Id: FileFeeder.java,v 1.16 2011/10/15 15:06:11 mal Exp $. now="+new java.util.Date());
	System.err.flush();
    }
}
