package com.polytechnik.itch;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;

/** ITCH 4.1 parser. Using
 * http://nasdaqtrader.com/content/technicalsupport/specifications/dataproducts/NQTV-ITCH-V4_1.pdf
 * specification.
 *  @version $Id: ITCHParser.java,v 1.242 2016/03/23 22:42:10 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class ITCHParser implements StreamProcessor {

    static final long
	SECOND_CONVERSION_SCALE=TraderInterface.TIME_SCALE,
	NANOSECOND_CONVERSION_SCALE=1;

    static final boolean PRG_DEBUG=true;

    // all null by default
    private final ITCHMessageProcessor [] processors=new ITCHMessageProcessor [MessageProcessor.MAX_N_CHAR_ARRAY_SIZE];
    // all 0 by default;
    private final int [] n_processed=new int [MessageProcessor.MAX_N_CHAR_ARRAY_SIZE];
    private int n_time_go_backward=0;

    private void addMessageProcessors(final ITCHMessageProcessor p)
    {
	final int t=p.getMessageType();
	if(t<0||t>=this.processors.length)
	{
	    throw new IllegalArgumentException("The processor type="+t+" ("+((char)t)+") is invalid. The character code should be between 0 and "+this.processors.length+".");
	}
	this.processors[t]=p;
    }
    @Deprecated
    public void overAll(final com.polytechnik.utils.TObjectProcedure<OrdersAndTrades> processor)
    {
	this.tradesbuilder.overAll(processor);
    }

    /** Uses standard BookBuilder & TradesBuilder.
     */
    public ITCHParser(final TraderInterface trader)
    {
	this(new BookBuilder(),trader);
    }
    /** A kludge to avod code before super().
     */
    private ITCHParser(final BookBuilder bookbuilder,
		       final TraderInterface trader)
    {
	this(bookbuilder,new TradesBuilder(bookbuilder),trader);
    }

    private final BookBuilder bookbuilder;
    final TradesBuilder tradesbuilder;
    private final TraderInterface trader;
    private long n_messages_received,n_bytes_processed;

    private /*public*/ ITCHParser(final BookBuilder bookbuilder,
		      final TradesBuilder tradesbuilder,
		      final TraderInterface trader)
    {
	this.bookbuilder=bookbuilder;
	this.tradesbuilder=tradesbuilder;
	this.trader=trader;
	this.n_messages_received=0;
	this.n_bytes_processed=0;

	// we always add these to handle standard actions
	addMessageProcessors(new SECONDMessageProcessor(this));
	addMessageProcessors(new SYSTEMEVENTMessageProcessor(this,this.trader));
	addMessageProcessors(new STOCKDIRECTORYMessageProcessor(this,this.trader));
	addMessageProcessors(new STOCKTRADINGACTIONMessageProcessor(this,this.trader));
	addMessageProcessors(new REGSHOACTIONMessageProcessor(this,this.trader));
	addMessageProcessors(new MARKETPARTICIPANTMessageProcessor(this,this.trader));
	addMessageProcessors(new ADDORDERMPIDMessageProcessor(this));
	addMessageProcessors(new ADDORDERMessageProcessor(this));
	addMessageProcessors(new ORDEREXECUTEDMessageProcessor(this));
	addMessageProcessors(new ORDEREXECUTEDWITHPRICEMessageProcessor(this));
	addMessageProcessors(new ORDERCANCELLEDMessageProcessor(this));
	addMessageProcessors(new ORDERDELETEMessageProcessor(this));
	addMessageProcessors(new REPLACEORDERMessageProcessor(this));
	addMessageProcessors(new TRADEMESSAGEMessageProcessor(this));
	addMessageProcessors(new CROSSTRADEMESSAGEMessageProcessor(this));
	addMessageProcessors(new BROKENTRADEMessageProcessor(this));
	addMessageProcessors(new NOIIMessageProcessor(this));
	addMessageProcessors(new RPIIMessageProcessor(this));
    }

    private long tsequence=200;
    private long getTradeSequence()
    {
	return ++this.tsequence;
    }



    long nseconds=0L,nanoseconds=0L;
    long getTime()
    {
	return this.nseconds+this.nanoseconds;
    }
    private String diagTimeHHMMSS()
    {
	return BookItem.getTimeDiag(getTime());
    }


    private static final char [] HEXCHARS=
	new char []{'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
    static final String getBytesAsString(final byte [] buf,final int b_start,final int b_len)
    {
	final StringBuilder sb=new StringBuilder(4*b_len+5);
	boolean was_hex=true;
	for(int i=b_start,n=b_len,i1=b_start+SoupBinTCP.MESSAGE_PADDING;n-->0;i++)
	{
	    final int c=buf[i];
	    if(' '<=c && c<127 && i>=i1)
	    {
		if(was_hex)
		{
		    sb.append(':');
		}
		sb.append((char)c);
		was_hex=false;
	    }
	    else
	    {
		if(!was_hex)
		{
		    sb.append(':');
		}
		sb.append("x");
		sb.append(HEXCHARS[(c&0xF0)>>>4]);
		sb.append(HEXCHARS[c&0x0F]);
		was_hex=true;
	    }
	}
	return sb.toString();
    }


    public int process(final byte [] buf,final int b_start,final int b_len)
    {
	if(b_len<SoupBinTCP.MESSAGE_PADDING)
	{
	    return 0; // empty string
	}
	// First two bytes are message length.
	final int m_length=MessageProcessor.getShortIntegerFromBytes(buf,b_start);
	if(m_length<=0)
	{
	    throw new IllegalStateException("m_length="+m_length);
	}
	final int lenp=m_length+SoupBinTCP.MESSAGE_PADDING;
	if(lenp>b_len)
	{
	    // to few data read, try next time
	    return 0;
	}

	final int mb_start=b_start+SoupBinTCP.MESSAGE_PADDING;
	// first character is message type
	final int t=buf[mb_start]&0xff;
	if(t<0||t>=MessageProcessor.MAX_N_CHAR_ARRAY_SIZE)
	{
	    throw new RuntimeException("The received message type="+t+" ("+((char)t)+") is invalid. The character code should be between 0 and "+this.processors.length+".");
	}
	this.n_processed[t]++;
	//System.err.println("TYPE="+((char)t));
	final ITCHMessageProcessor p=this.processors[t];
	if(p==null)
	{
	    throw new RuntimeException("Unknown message type="+t+" ("+((char)t)+") for string "+getBytesAsString(buf,b_start,Math.min(b_len,1000)));
	}
	if(false && p.getMessageType()!=t) // no need to validate here
	{
	    throw new IllegalArgumentException(
		"The string =`"+getBytesAsString(buf,b_start,Math.min(b_len,1000))+
		"' has type="+p.getMessageType()+" ("+((char)p.getMessageType())+
		") wrong. Should be "+t+" ("+((char)t)+").");
	}
	final int len=p.getLength();
	if(len!=m_length)
	{
	   throw new IllegalArgumentException(
		    "The string =`"+getBytesAsString(buf,b_start,Math.min(b_len,1000))+
		    "' is invalid length mismatch lenread="+m_length+
		    " len="+len+" for messages of "+((char)t)+" type.");
	}
	if(t!='T')
	{ // for all but "SECOND" message extract nanoseconds.
	    if(len<5)
	    {
		throw new IllegalArgumentException(
		    "The string =`"+getBytesAsString(buf,b_start,Math.min(b_len,1000))+
		    "' is invalid len="+len+"<5; b_len="+b_len+
		    ". There should be "+(SoupBinTCP.MESSAGE_PADDING+1+4)+
		    " bytes: (padding(2), type(1) and nanoseconds(4) in the beginning");
	    }
	    // offset=1 (message type, then read next 4 bytes nanoseconds).
	    final long onanoseconds=this.nanoseconds;
	    // Seconds change are handled separately in SECONDMessageProcessor
	    // where nanoseconds are set to 0.
	    if((this.nanoseconds=ITCHParser.NANOSECOND_CONVERSION_SCALE*MessageProcessor.getIntegerFromBytes(buf,mb_start+1))<onanoseconds)
	    {
		System.err.println(
		    // Do not die, this happen relatively seldom.
		    // throw new IllegalArgumentException(
		    "!!!!!!DATA ERROR: Message Type="+((char)t)+" len=2+"+len+
		    ". Time go backward; received nanoseconds="+
		    this.nanoseconds+" < "+onanoseconds+". The input string =`"+
		    getBytesAsString(buf,b_start,lenp)+"'");
		this.nanoseconds=onanoseconds;
		this.n_time_go_backward++;
	    }
	}
	p.processMessage(buf,mb_start);

	/*
	// debug
	if(true)
	{
	    final long t0=getTime(),dt=2*TraderInterface.TIME_SCALE;
	    if(Math.abs(36185243901112L-t0)<dt||
	       Math.abs(36176619411661L-t0)<dt||
	       Math.abs(36159010102300L-t0)<dt||
	       Math.abs(36156850508948L-t0)<dt||
	       Math.abs(36148537716195L-t0)<dt||
	       Math.abs(36142887031480L-t0)<dt)
	    {
		System.err.println("DIAG: The input t0="+t0+" string =`"+
				   getBytesAsString(buf,b_start,lenp)+"' p="+p);
	    }
	}
	*/

	this.n_messages_received++;
	this.n_bytes_processed+=lenp;
	if(PRG_DEBUG)
	{
	    if(this.n_messages_received%100000==0)
	    {
		System.err.println("ITCHParser:: "+this.n_messages_received+" messages; "+
				   this.n_bytes_processed+" bytes processed; itch_time="+
				   diagTimeHHMMSS()+" now="+new java.util.Date());
	    }
	}
	return lenp;
    }


    private final java.util.HashMap<String,String> stringeliminator=new java.util.HashMap<String,String>(4001,0.27f);
    /** A kludgy way to eliminate multiple instances of strings.
     */
    final String eliminateMultipleInstances(final String s)
    {
	if(s==null)
	{
	    return null;
	}
	final String s2=this.stringeliminator.get(s);
	if(s2!=null)
	{
	    return s2;
	}
	this.stringeliminator.put(s,s);
	return s;
    }



    public final String getStats()
    {
	final StringBuilder buf=new StringBuilder();
	buf.append("============= BOOK STAT AT T="+getTime()+
		   " "+diagTimeHHMMSS()+" ver $Id: ITCHParser.java,v 1.242 2016/03/23 22:42:10 mal Exp $");
	buf.append('\n');
	buf.append("BookBuilder & TradesBuilder STAT AT T="+getTime()+
		   " "+diagTimeHHMMSS()+"\n");
	buf.append(this.tradesbuilder.getStats());
	buf.append("Processors STAT AT T="+getTime()+
		   " "+diagTimeHHMMSS()+"\n");
	for(int i=0;i<this.processors.length;i++)
	{
	    final ITCHMessageProcessor p=this.processors[i];
	    if(p!=null)
	    {
		buf.append(this.n_processed[i]);
		buf.append(":records:");
		buf.append((char)p.getMessageType());
		buf.append(':');
		buf.append(p.getClass().getName());
		buf.append('\n');
	    }
	}
	buf.append("N time go backward=");
	buf.append(this.n_time_go_backward);
	buf.append('\n');
	return buf.toString();
    }

    private boolean printDiag(final OrdersAndTrades ot,final long currenttime)
    {
	/*
	if(ot!=null && ot.ticker.equals("NXPI"))
	{
	    return true;
	}
	*/

	/*
	if(ot!=null && ot.ticker.equals("AAPL") &&
	   currenttime>(9*3600+40*60)*TraderInterface.TIME_SCALE)
	{
	    return true;
	}
	*/
	return false;
    }



// =========== implementation of methods that distribure read data to trading module ========


void addOrder(final ADDORDERMPIDMessageProcessor.ADDORDERMPIDMessage m)
{
    final OrdersAndTrades ot=this.tradesbuilder.initTicker(m.stock);
    if(true) //synchronized(ot.ticker)
    {
	final byte buy_sell_indicator=MessageProcessor.getBuySellExternal(m.buy_sell_indicator);
	final BookItem b=new BookItem(m.price,
				      m.shares,
				      this.eliminateMultipleInstances(m.attribution),
				      buy_sell_indicator,
				      m.time,
				      m.orderreferencenumber,
				      (buy_sell_indicator==BookItem.BOOK_BUY?ot.orderbook.getBestOrder(BookItem.BOOK_SELL):(buy_sell_indicator==BookItem.BOOK_SELL?ot.orderbook.getBestOrder(BookItem.BOOK_BUY):null)));
	this.bookbuilder.addNewItem(m.orderreferencenumber,ot.orderbook,b);
	processPseudoTrades(ot,m.time,null);
	this.trader.update(ot,m.time,TraderInterface.UPD_ADD_ORDER);
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+ot.ticker+":addOrder:"+m.dumpData());
    }
}

void executeOrder(final ORDEREXECUTEDWITHPRICEMessageProcessor.ORDEREXECUTEDWITHPRICEMessage m)
{
    updateOrder(m.orderreferencenumber,m.shares,m.time,m);
}

private void updateOrder(final long orderreferencenumber,
			 final int shares,
			 final long t_now,
			 final ORDEREXECUTEDWITHPRICEMessageProcessor.ORDEREXECUTEDWITHPRICEMessage m)
{
    final RealtimeBook tickerbook=this.bookbuilder.idLookupBook(orderreferencenumber);
    final OrdersAndTrades ot;
    if(tickerbook!=null)
    {
	ot=tickerbook.ref;
	if(true) //synchronized(ot.ticker)
	{
	    final BookItem o=tickerbook.get(orderreferencenumber);
	    if(o==null && m==null)
	    {
		if(PRG_DEBUG)
		{
		    System.err.println("ITCHParser: a cancel of already cancelled order="+
				       orderreferencenumber+" shares="+shares+
				       " ticker="+ot.ticker+
				       " t_now="+t_now+" ("+diagTimeHHMMSS()+")"+
				       " m="+m);
		}
		return;
	    }
	    if(o==null)
	    {
		// execute non-existent order
		throw new RuntimeException("The order="+orderreferencenumber+
					   " for "+ot.ticker+" is not found.");
	    }

	    final BookItem biu;
	    final int new_shares=o.shares-shares;
	    if(new_shares<=0)
	    {
		// should we remove the order from the book?
		biu=this.bookbuilder.removeItem(orderreferencenumber,tickerbook);
	    }
	    else
	    {
		// replace the order by one with lower number of shares.
		// We consider order time changed only in case of
		// partial execution (m!=null).
		// In case of partical cancelling (m=null)
		// we do not adjust order time.
		// Let me think more about what time to use when original order is placed.
		// Oct 21 2008.
		// Rollback. Set t=t_now in both cases (partial execution & partical cancelling)
		// Oct 22 2008
		// set logic back.

		// Feb 20 2008. Order time never changes.
		// This will lead to failures of some previous unit tests.
		// Create new data.
		biu=tickerbook.put(orderreferencenumber,
				   BookItem.adjustShares(o,new_shares,/* t_now */ /*  m!=null?t_now: */o.time));
		// no need to guess origtime, because time & id did not change
	    }

	    if(o!=biu)
	    {
		throw new IllegalStateException();
	    }

	    if(m!=null)
	    {
		// Execution/partial execution
		final Trade t;
		final double price=m.executionprice>0.0?m.executionprice:o.price;
		this.tradesbuilder.addNewTrade(
		    ot.tradebook,
		    t=new Trade(price,
				m.shares,
				o.attribution,
				o.buy_sell_indicator,
				m.time,
				m.matchnumber,
				m.orderreferencenumber,
				o.time,
				biu,
				Trade.TRADE_TYPE_REGULAR,
				OrdersAndTrades.SAVE_EDGE_ORDERS?tickerbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_BUY):null,
				OrdersAndTrades.SAVE_EDGE_ORDERS?tickerbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_SELL):null,
				getTradeSequence()));
		// Executed trades not considered as cancelled.
		// No no need to modify cancelled orders list.

		if(printDiag(ot,t_now))
		{
		    System.err.println("DIAG:"+ot.ticker+":updateOrder:"+m.dumpData()+
				       " trade="+Trade.getTradeDescription(t));
		}

		processPseudoTrades(ot,t_now,t);
		// Notify trader with updated realtime book & trades.
		this.trader.update(ot,t_now,TraderInterface.UPD_EXECUTIONORDER_TRADE);
	    }
	    else
	    {
		// m==null means we have order cancel.

		processPseudoTrades(ot,t_now,null);
		// Partial order canceling.
		// Notify trader with updated realtime book and the same trades.
		this.trader.update(ot,t_now,TraderInterface.UPD_CANCEL_ORDER);
	    }
	    if(new_shares<0)
	    {
		throw new RuntimeException("The order="+orderreferencenumber+
					   " for "+ot.ticker+
					   " is adjusted to negative number of shares="+new_shares);
	    }
	}
    }
    else
    {
	ot=null;
	if(PRG_DEBUG)
	{
	    System.err.println("ITCHParser: an update of already cancelled order="+
			       orderreferencenumber+" shares="+shares+
			       " t_now="+t_now+" ("+diagTimeHHMMSS()+")"+
			       " m="+m);
	}
    }

    if(printDiag(ot,t_now))
    {
	System.err.println("DIAG:"+(ot!=null?ot.ticker:"????")+":updateOrder:"+
			   (m!=null?m.dumpData():("cancel orderreferencenumber="+orderreferencenumber+" shares="+shares)));
    }
}



void replaceOrder(final REPLACEORDERMessageProcessor.REPLACEORDERMessage m)
{
    final RealtimeBook tickerbook=this.bookbuilder.idLookupBook(m.original_orderreferencenumber);
    if(tickerbook==null)
    {
	throw new RuntimeException("The order="+m.original_orderreferencenumber+
				   " is not found");
    }
    final OrdersAndTrades ot=tickerbook.ref;
    if(true) //synchronized(ot.ticker)
    {
	final BookItem o=this.bookbuilder.removeItem(m.original_orderreferencenumber,tickerbook);
	if(o==null)
	{
	    // replace non-existent order
	    throw new RuntimeException("The order="+m.original_orderreferencenumber+
				       " for "+ot.ticker+" is not found.");
	}
	// put new order to the book
	final BookItem b=new BookItem(m.price,
				      m.shares,
				      o.attribution,
				      o.buy_sell_indicator,
				      m.time,
				      m.new_orderreferencenumber,
				      (o.buy_sell_indicator==BookItem.BOOK_BUY?tickerbook.getBestOrder(BookItem.BOOK_SELL):(o.buy_sell_indicator==BookItem.BOOK_SELL?tickerbook.getBestOrder(BookItem.BOOK_BUY):null)));

	this.bookbuilder.addNewItem(m.new_orderreferencenumber,tickerbook,b);
	processPseudoTrades(ot,m.time,null);
	this.trader.update(ot,m.time,TraderInterface.UPD_CANCELREPLACE_ORDER);
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+ot.ticker+":replaceOrder:"+m.dumpData());
    }
}


void cancelOrder(final ORDERCANCELLEDMessageProcessor.ORDERCANCELLEDMessage m)
{
    updateOrder(m.orderreferencenumber,m.shares,m.time,null);
}

void deleteOrder(final ORDERDELETEMessageProcessor.ORDERDELETEMessage m)
{
    final RealtimeBook tickerbook=this.bookbuilder.idLookupBook(m.orderreferencenumber);
    final OrdersAndTrades ot;
    if(tickerbook==null)
    {
	ot=null;
	if(PRG_DEBUG)
	{
	    System.err.println("ITCHParser: a delete of already deleted order="+
			       m.orderreferencenumber+
			       " t_now="+m.time+" ("+diagTimeHHMMSS()+")");
	}
    }
    else
    {
	ot=tickerbook.ref;
	if(true) //synchronized(ot.ticker)
	{
	    final BookItem biu=this.bookbuilder.removeItem(m.orderreferencenumber,tickerbook);
	    if(biu!=null)
	    {
		processPseudoTrades(ot,m.time,null);
		// update notification
		this.trader.update(ot,m.time,TraderInterface.UPD_CANCEL_ORDER);
	    }
	    else
	    {
		if(PRG_DEBUG)
		{
		    System.err.println("ITCHParser: a delete of already deleted order="+
				       m.orderreferencenumber+
				       " ticker="+ot.ticker+
				       " t_now="+m.time+" ("+diagTimeHHMMSS()+")");
		}
	    }
	}
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+(ot!=null?ot.ticker:"????")+":deleteOrder:"+m.dumpData());
    }
}


void executeNonDisplayableOrder(final TRADEMESSAGEMessageProcessor.TRADEMESSAGEMessage m)
{
    if(!OrdersAndTrades.SAVE_NONDISPLAYABLE_TRADES)
    {
	return;
    }
    final OrdersAndTrades ot=this.tradesbuilder.initTicker(m.stock);
    if(true) //synchronized(ot.ticker)
    {
	final Trade t;
	// Notify trader with the same realtime book & updated trades.
	this.tradesbuilder.addNewTrade(
	    ot.tradebook,
	    t=new Trade(m.price,
			m.shares,
			null,
			MessageProcessor.getBuySellExternal(m.buy_sell_indicator),
			m.time,
			m.matchnumber,
			m.orderreferencenumber,
			this.bookbuilder.guessTimeFromID(m.orderreferencenumber,m.time),
			null,
			Trade.TRADE_TYPE_NONDISPLAYABLE,
			OrdersAndTrades.SAVE_EDGE_ORDERS?ot.orderbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_BUY):null,
			OrdersAndTrades.SAVE_EDGE_ORDERS?ot.orderbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_SELL):null,
			getTradeSequence()));

	processPseudoTrades(ot,m.time,t);
	this.trader.update(ot,m.time,TraderInterface.UPD_EXECUTIONOTH_TRADE);
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+ot.ticker+":executeNonDisplayableOrder:"+m.dumpData());
    }
}


void executeCrossTrade(final CROSSTRADEMESSAGEMessageProcessor.CROSSTRADEMESSAGEMessage m)
{
    if(!OrdersAndTrades.SAVE_NONDISPLAYABLE_TRADES)
    {
	return;
    }
    final OrdersAndTrades ot=this.tradesbuilder.initTicker(m.stock);
    if(true) //synchronized(ot.ticker)
    {
	final Trade t;
	this.tradesbuilder.addNewTrade(
	    ot.tradebook,
	    t=new Trade(m.price,
		      (int)m.shares, // if it wrap - we do not actually care, because we actually do not handle BookItem.BOOK_UNKNOWN
		      null,
		      BookItem.BOOK_UNKNOWN,
		      m.time,
		      m.matchnumber,
		      0L, // cross has no order reference
		      -1, // cross has no origtime
		      null,
		      Trade.TRADE_TYPE_CROSS,
		      OrdersAndTrades.SAVE_EDGE_ORDERS?ot.orderbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_BUY):null,
		      OrdersAndTrades.SAVE_EDGE_ORDERS?ot.orderbook.getBestOrder(/*getBestLevel(m.time,*/BookItem.BOOK_SELL):null,
		      getTradeSequence()));
	processPseudoTrades(ot,m.time,t);
	// Notify trader with the same realtime book & updated trades.
	this.trader.update(ot,m.time,TraderInterface.UPD_EXECUTIONOTH_TRADE);
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+ot.ticker+":executeCrossTrade:"+m.dumpData());
    }
}


void processBrokenTrade(final BROKENTRADEMessageProcessor.BROKENTRADEMessage m)
{

    final TradeList tl=this.tradesbuilder.idLookupTrades(m.matchnumber);
    final OrdersAndTrades ot;
    if(tl!=null)
    {
	ot=tl.ref;
	if(true) //synchronized(ot.ticker)
	{
	    final TradeList tradelist=this.tradesbuilder.removeTrade(m.matchnumber);
	    if(tradelist!=null)
	    {
		if(tradelist.ref!=ot)
		{
		    throw new IllegalStateException();
		}
		processPseudoTrades(ot,m.time,null);
		// trade found and removed.
		// Notify trader with the same realtime book & updated trades.
		this.trader.update(ot,m.time,TraderInterface.UPD_BROKENTRADE_TRADE);
	    }
	}
    }
    else
    {
	ot=null;
    }
    if(printDiag(ot,m.time))
    {
	System.err.println("DIAG:"+(ot!=null?ot.ticker:"????")+":processBrokenTrade:"+m.dumpData());
    }
}


void processPseudoTrades(final OrdersAndTrades ot,final long currenttime,final Trade trade)
{
    if(!ot.FLAG_PSEUDOTRADES||(!ot.getActive())) return;
    if(trade!=null)
    { // true trade
	ot.pseudotrades.add(trade);
	return;
    }
    final Trade tl=ot.tradebook.getLastAddedElement();
    if(tl==null) return;

    // pseudotrade vol=0
    ot.pseudotrades.add(
	new Trade(tl.price,
		  0, // 0 volume for pseudotrades
		  null,
		  tl.buy_sell_indicator,
		  currenttime,
		  0,
		  0,
		  0,
		  null,
		  Trade.TRADE_TYPE_PSEUDO,
		  ot.orderbook.getBestOrder(/*getBestLevel(currenttime,*/BookItem.BOOK_BUY),
		  ot.orderbook.getBestOrder(/*getBestLevel(currenttime,*/BookItem.BOOK_SELL),
		  0));
}



// =========== END implementation of methods that distribure read data to trading module ========

}
