package com.polytechnik.itch;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */


/** An interface to process message.
 *  @version $Id: MessageProcessor.java,v 1.49 2012/08/15 20:12:20 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
abstract class MessageProcessor implements ITCHMessageProcessor {


    static final int MAX_N_CHAR_ARRAY_SIZE=128;

    /** Generic constants.
     */
    static final int G_YES='Y',G_NO='N',G_BUY='B',G_SELL='S';
    // all false by default
    static final boolean []
	validate_buy_sell_indicator=new boolean [MAX_N_CHAR_ARRAY_SIZE],
	validate_yes_no=new boolean [MAX_N_CHAR_ARRAY_SIZE];

    static {

	validate_buy_sell_indicator[G_BUY]=true;
	validate_buy_sell_indicator[G_SELL]=true;

	validate_yes_no[G_YES]=true;
        validate_yes_no[G_NO]=true;
    }


    private final int type;
    final int [] offsets;
    final ITCHParser parser;
    MessageProcessor(final int type,
		     final int [] offsets,
		     final ITCHParser parser)
    {
	this.type=type;
	this.offsets=offsets;
	this.parser=parser;

    }

    public int getMessageType()
    {
	return this.type;
    }

    public int getLength()
    {
	return this.offsets[this.offsets.length-1];
    }

    static byte getBuySellExternal(final int itch_buy_sell)
    {
	if(itch_buy_sell==G_BUY)
	{
	    return com.polytechnik.realtimebook.BookItem.BOOK_BUY;
	}
	if(itch_buy_sell==G_SELL)
	{
	    return com.polytechnik.realtimebook.BookItem.BOOK_SELL;
	}
	return com.polytechnik.realtimebook.BookItem.BOOK_UNKNOWN;
    }

    static long getLongFromBytes(final byte [] buf,final int f)
    {
	return ((buf[f]&0xffL) <<56) |
	    ((buf[f+1]&0xffL) <<48) |
	    ((buf[f+2]&0xffL) <<40) |
	    ((buf[f+3]&0xffL) <<32) |
	    ((buf[f+4]&0xffL) <<24) |
	    ((buf[f+5]&0xffL) <<16) |
	    ((buf[f+6]&0xffL) <<8) |
	    (buf[f+7]&0xffL);
    }

    long getLong(final byte [] buf,final int b_start,final int field_num)
    {
	final int o=this.offsets[field_num-1];
	if(o+8!=this.offsets[field_num])
	{
	    throw new RuntimeException("Invalid field length="+
				       (this.offsets[field_num]-o)+
				       " must be 8 for the string=`"+
				       getOriginalString(buf,b_start)+"'.");
	}
	final long i=getLongFromBytes(buf,b_start+o);

	if(i<0)
	{
	    throw new RuntimeException("Invalid (negtive) data read for the string =`"+
				       getOriginalString(buf,b_start)+"'. Data="+i);
	}
	return i;
    }

    int getInteger(final byte [] buf,final int b_start,final int field_num)
    {
	final int o=this.offsets[field_num-1];
	if(o+4!=this.offsets[field_num])
	{
	    throw new RuntimeException("Invalid field length="+
				       (this.offsets[field_num]-o)+
				       " must be 4 for the string=`"+
				       getOriginalString(buf,b_start)+"'.");
	}
	final int i=getIntegerFromBytes(buf,b_start+o);

	// there are no negative numbers here
	if(i<0)
	{
	    throw new RuntimeException("Invalid (negtive) data read for the string =`"+
				       getOriginalString(buf,b_start)+"'. Data="+i);
	}
	return i;
    }

    static int getIntegerFromBytes(final byte [] buf,final int f)
    {
	return
	    ((buf[f]&0xff) <<24) |
	    ((buf[f+1]&0xff) <<16) |
	    ((buf[f+2]&0xff) <<8) |
	    (buf[f+3]&0xff);
    }

    static int getShortIntegerFromBytes(final byte [] buf,final int f)
    {
	return ((buf[f]&0xff) <<8) | (buf[f+1]&0xff);
    }

    private static String getStringFromBytes(final byte [] buf,final int o,int o_end)
    {
	if(o>=o_end)
	{
	    // treat empty by definition string as null
	    return null;
	}
	// all actual strings are right padded.
	for(;o<o_end && buf[o_end-1]==' ';o_end--);
	return getStringFromBytesOfLen(buf,o,o_end-o);
    }

    static String getStringFromBytesOfLen(final byte [] buf,final int o,int len)
    {
	if(len<=0)
	{
	    // treat empty by definition string as null
	    return null;
	}
	// use deprecated String constructor, it is substantially faster
	@SuppressWarnings("deprecation")
	final String sr=new String(buf,0,o,len);
	return sr;
    }

    String getString(final byte [] buf,final int b_start,final int field_num)
    {
	return getStringFromBytes(buf,b_start+this.offsets[field_num-1],b_start+this.offsets[field_num]);
    }


    double getPrice(final byte [] buf,final int b_start,final int field_num)
    {
	return ((double)getInteger(buf,b_start,field_num))/10000.0;
    }

    int getChar(final byte [] buf,final int b_start,final int field_num)
    {
	final int o=this.offsets[field_num-1];
	if(o+1!=this.offsets[field_num])
	{
	    throw new RuntimeException("Invalid field length="+
				       (this.offsets[field_num]-o)+
				       " must be 1 for the string=`"+
				       getOriginalString(buf,b_start)+"'.");
	}
	return buf[b_start+o];
    }

    void validateChar(final int c,
		      final boolean [] allowed_values,
		      final String descr,
		      final byte [] buf,
		      final int b_start)
    {
	if(c<0||c>=allowed_values.length || !allowed_values[c])
	{
	    throw new IllegalArgumentException(
		"The string =`"+getOriginalString(buf,b_start)+
		"' has wrong "+descr+" value="+((char)c)+
		" for message of type="+getMessageType()+
		" ("+((char)getMessageType())+").");
	}
    }


    private String getOriginalString(final byte [] buf,final int b_start)
    {
	return ITCHParser.getBytesAsString(buf,b_start,getLength());
    }

}
