package com.polytechnik.realtimebook;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */


/** A class to store orders flow.
 *  @version $Id: OrdersFlow.java,v 1.7 2009/06/27 00:20:53 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class OrdersFlow
{

    /* private */ double [] price;
    /* private */ long [] vol_total;
    private int n_elements;
    private long [] shares;
    final boolean flag_descending;

    public OrdersFlow(final boolean flag_descending)
    {
	this(100,flag_descending);
    }

    OrdersFlow(final int initialsize,final boolean flag_descending)
    {
	this.price=new double [initialsize];
	this.shares=new long [initialsize];
	this.vol_total=new long [initialsize];
	this.n_elements=0;
	this.flag_descending=flag_descending;
    }

    public void clear()
    {
	this.n_elements=0;
	java.util.Arrays.fill(this.price,Double.NaN);
	java.util.Arrays.fill(this.shares,Long.MIN_VALUE);
	java.util.Arrays.fill(this.vol_total,Long.MIN_VALUE);
    }

    public int getSize()
    {
	return this.n_elements;
    }


    private void ensureCapacity(final int newsize)
    {
	if(newsize<=this.price.length)
	{
	    throw new IllegalArgumentException("Try to catch wrong usage");
	}
	final double [] newprice=new double [newsize];
	System.arraycopy(this.price,0,newprice,0,this.n_elements);
	this.price=newprice;
	final long [] newshares=new long [newsize];
	System.arraycopy(this.shares,0,newshares,0,this.n_elements);
	this.shares=newshares;
	final long [] newvol_total=new long [newsize];
	// no need to copy the total volume data
	this.vol_total=newvol_total;
    }


    public int addOrder(final double price,final long shares)
    {
	if(Double.isNaN(price)||shares<=0)
	{
	    throw new IllegalStateException("Negative or zero volume shares="+shares+" price="+price);
	}
	final int n=this.n_elements;
	if(n>=this.price.length)
	{
	    ensureCapacity(2*n+1);
	}
	this.price[n]=price;
	this.shares[n]=shares;
	this.vol_total[n]=Long.MIN_VALUE; // to catch misusage
	return ++this.n_elements;
    }

    public int addOrderCheckPrev(final double price,final long shares)
    {
	if(Double.isNaN(price)||shares<=0)
	{
	    throw new IllegalStateException("Negative or zero volume shares="+shares+" price="+price);
	}
	final int n1=this.n_elements-1;
	if(n1<0||Math.abs(this.price[n1]-price)>OrdersFlow.EPS)
	{
	    return addOrder(price,shares);
	}
	// same price, add to existing element volume
	this.shares[n1]+=shares;
	this.vol_total[n1]=Long.MIN_VALUE; // to catch misusage
	return this.n_elements;
    }


    /** Get orders from other flow.
     *  @param ord Source.
     */
    public int addOrdersFrom(final OrdersFlow ord)
    {
	if(false && this.flag_descending!=ord.flag_descending)
	{
	    throw new IllegalArgumentException();
	}
	final int n0=getSize(),n1=ord.getSize();
	final int fsize=n0+n1;
	if(fsize>this.price.length)
	{
	    ensureCapacity(2*fsize+1);
	}
	System.arraycopy(ord.price,0,this.price,n0,n1);
	System.arraycopy(ord.shares,0,this.shares,n0,n1);
	return (this.n_elements=fsize);
    }

    public void dump(final String txt)
    {
	System.err.println("DUMP="+txt+"\t type="+this.flag_descending+
			   " class="+this.getClass().getName());
	System.err.println("N\tPrice\tshares\t");
	for(int i=0;i<this.n_elements;i++)
	{
	    System.err.println(String.valueOf(i)+"\t"+
			       this.price[i]+"\t"+
			       this.shares[i]);

	}
    }


    public int mergeIdenticalPrices()
    {
	final int n=getSize();
	if(n<=0)
	{
	    return 0;
	}
	final long [] s=this.shares;
	final double [] p=this.price;
	com.polytechnik.utils.TMPQSortAlgorithm.doubleQuickSortLongLoad(
	    p,
	    0,
	    n-1,
	    this.flag_descending,
	    s);
	if(s[0]<=0||Double.isNaN(p[0]))
	{
	    throw new IllegalStateException("Negative or zero volume="+s[0]+
					    " or invalid price="+p[0]);
	}

	double pip=p[0],pprev=pip;
	int ip=0;
	for(int i=1;i<n;i++)
	{
	    final double pi=p[i];
	    if(s[i]<=0||Double.isNaN(pi))
	    {
		throw new IllegalStateException("Negative or zero volume["+i+"]="+s[i]+
						" or price["+i+"]="+pi);
	    }
	    if(this.flag_descending?pi>pprev:pi<pprev)
	    {
		throw new IllegalStateException("Problem with sorting");
	    }
	    if(Math.abs(pi-pip)>OrdersFlow.EPS)
	    {
		pip=p[++ip]=pi;
		s[ip]=s[i];
	    }
	    else
	    {
		if((s[ip]+=s[i])<=0)
		{
		    throw new IllegalStateException("Overflow");
		}
	    }

	}
	java.util.Arrays.fill(this.vol_total,Long.MIN_VALUE); // invalidate to catch usage errors
	return (this.n_elements=ip+1);
    }


    public long getTotalVolume()
    {
	long v=0;
	final boolean fd=this.flag_descending;
	double pprev=Double.NaN;
	final double [] p=this.price;
	final long [] s=this.shares;
	final long [] vtot=this.vol_total;
	for(int i=0,n=this.n_elements;i<n;i++)
	{
	    if(fd?p[i]>pprev:p[i]<pprev)
	    {
		throw new RuntimeException(
		    "data is not sorted: price["+i+"]="+p[i]+
		    " pprev="+pprev);
	    }
	    pprev=p[i];
	    vtot[i]=(v+=s[i]);
	    if(v<=0 || Double.isNaN(pprev))
	    {
		throw new IllegalStateException(
		    "overflow vol["+i+"]="+v+
		    " or invalid price["+i+"]="+pprev);
	    }
	}
	return v;
    }


    /** Shift all prices on specific value.
     */
    public int shiftPrices(final double shift)
    {
	if(Double.isNaN(shift))
	{
	    throw new IllegalArgumentException("shift="+shift);
	}
	final double [] p=this.price;
	for(int i=this.n_elements;i-->0;)
	{
	    if(Double.isNaN(p[i]+=shift))
	    {
		throw new IllegalArgumentException("price["+i+"]="+p[i]);
	    }
	}
	return this.n_elements;
    }

    /** Shift & scale all prices on specific value.
     */
    public int shiftAndScalePrices(final double shift,final double scale)
    {
	if(Double.isNaN(shift)||Double.isNaN(scale))
	{
	    throw new IllegalArgumentException("shift="+shift+" scale="+scale);
	}
	final double [] p=this.price;
	for(int i=this.n_elements;i-->0;)
	{
	    if(Double.isNaN(p[i]=p[i]*scale+shift))
	    {
		throw new IllegalArgumentException("price["+i+"]="+p[i]);
	    }
	}
	return this.n_elements;
    }



    public static class DistStat {
	public final double median,min,max;
	public final long vtotal;
	public final int np;

	DistStat(final double median,
		 final double min,
		 final double max,
		 final long vtotal,
		 final int np)
	{
	    this.median=median;
	    this.min=min;
	    this.max=max;
	    this.vtotal=vtotal;
	    this.np=np;
	}
    }

    /** Calculate min,max,median & volume.
     */
    public DistStat getStat(final double frac)
    {
	mergeIdenticalPrices();

	final long v0=getTotalVolume(),lev=(long)(frac*v0);
	// vol_total is set above in getTotalVolume()
	final long [] vtot=this.vol_total;
	final double [] p=this.price;
	final int n=this.n_elements;

	double pprev=Double.NaN;
	for(int i=0;i<n;i++)
	{
	    final double pi=p[i];
	    if(this.flag_descending?pi>pprev:pi<pprev)
	    {
		throw new IllegalStateException("Problem with sorting");
	    }
	    pprev=pi;
	    if(vtot[i]>lev)
	    {
		break;
	    }
	}
	return new DistStat(pprev,
			    n>0?p[this.flag_descending?n-1:0]:Double.NaN,
			    n>0?p[this.flag_descending?0:n-1]:Double.NaN,
			    v0,
			    n);
    }


    /** Price roundoff. */
    public static final double EPS=1e-6;


    /** For debugging. */
    public static interface IterateThrough {
	boolean execute(final double price,final long shares);
    }
    /** For debugging. */
    public boolean forEachEntry(IterateThrough  procedure)
    {
	for(int i=0,n=getSize();i<n;i++)
	{
	    if(!procedure.execute(price[i],shares[i]))
	    {
		return false;
	    }
	}
	return true;
    }

}
