package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Basis of orthogonal polynomials.
 *  @version $Id: BasisPolynomials.java,v 1.24 2016/03/10 21:10:44 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class BasisPolynomials {

/** Calculates n-th basis polynomial (in monomials basis).
 *  This mononials basis calculation is used for unit test only.
 * @param n T index.
 * @param res Output vector.
 * @param Tnm1 Input T_{n-1}
 * @param Tnm2 Input T_{n-2}
 */
public void setNext(final int n,final double [] res,
		    final double [] Tnm1,final double [] Tnm2)
{
    setNext(n,res,Tnm1,Tnm2,false);
}
public void setNextType2(final int n,final double [] res,
			 final double [] Tnm1,final double [] Tnm2)
{
    setNext(n,res,Tnm1,Tnm2,true);
}
protected abstract void setNext(final int nd,final double [] res,
				final double [] Tnm1,final double [] Tnm2,
				final boolean flag_type2);


/**
 * Convert from basis to polynomial basis. Used for testing.
 * @param n T index.
 * @param Tcoefs Coefficients of T_n.
 * @return Polynomial in regular basis.
 */
public double [] convertToPolynomialFromBasis_Test(final int n,final double [] Tcoefs)
{
    double [] ra=new double [n+1],rb=new double [n+1],sum=new double [n+1];
    for(int k=0;k<=n;k++)
    {
	setNext(k,rb,ra,rb);
	final double [] l=rb;rb=ra;ra=l;
	TMatr.sumP1(sum,l,k+1,Tcoefs[k]);
    }
    return sum;
}
/**
 * Convert from polynomials to basis. Used for testing.
 * Note: may be numerically unstable.
 * @param n max power.
 * @param coefs Polynomial coefficients.
 * @return Polynomial in basis.
 */
public double [] convertToBasisFromPolynomial_Test(final int n,final double [] coefs)
{
    final double [][] P=new double [n+1][];
    for(int k=0;k<=n;k++)
    {
	final double [] pk=P[k]=new double [k+1];
	setNext(k,pk,k<1?null:P[k-1],k<2?null:P[k-2]);
    }
    final double [] b=new double [n+1];
    TMatr.expandOverLowerDiagBasis(b,coefs,n+1,P);
    return b;
}


/**
 * Calculation of the value of the basis polynomial.
 * @param n T index.
 * @param x Argument.
 * @return The value of the bais polynomial Tn at x.
 */
public abstract double calculate(final int n,final double x);


/** Given observations x and weights w calculate the 0..n moments in the basis.
 * k-th element is \sum_{j=0}^{j=x.length-1} T_k(x[j])*w[j].
 */
public abstract double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w,final int nelsinsample);
public double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w)
{
    if(w!=null && w.length!=x.length) throw new IllegalArgumentException();
    return calculateMomentsFromSample(n,x,w,x.length);
}
public double [] calculateMomentsFromSample(final int n,final double [] x)
{
    return calculateMomentsFromSample(n,x,null);
}


/** Calculates sum_{k=0}^{k=n}T_k(x) Tcoefs_k.
 * @param n T index.
 * @param Tcoefs Coefficients of T_n.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public abstract double calculateSumD0(final int n,final double [] Tcoefs,final double x);


/** Calculates d/dx(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public abstract double calculateSumD1(final int n,final double [] Tcoefs,final double x);


/** Calculates d^2/dx^2(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public abstract double calculateSumD2(final int n,final double [] Tcoefs,final double x);



/** Differentiate a function in basis,
 *  F(x) - a polynomial of n-th order.
 *  @param n Max index.
 *  @param Tcoefs: F(x)=T_k(x)*Tcoefs[k].
 *  @return dF/dx=T'_k(x)*Tcoefs[k], converted to basis.
 */
public abstract double [] getDifferentiated(final int n,final double [] Tcoefs);

/** Integrate a function in basis.
 *  F(x) - a polynomial of n-th order.
 *  @param n Max index.
 *  @param Tcoefs, such that F(x)=T_k(x)*Tcoefs[k].
 *  @return \int F(x)dx=Tcoefs[k]*\int T_k(x) dx, converted to basis.
 *  The integral is typically normalized as I(0)=0.
 */
public abstract double [] getIntegrated(final int n,final double [] Tcoefs);



/** Expand Tl(z)Tm(z)=sum_k=0^{l+m} s_k Tk(z),
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public abstract void setQlQmExpansionCoefs(final int l,final int m,final double [] s);


/** Multiply two polynomials in basis.
 * @param res The result, first n1+n2+1 elements is set.
 * @param p1 Firts polynomial.
 * @param n1 Max index of p1, the dimension of p1 is n1+1.
 * @param p2 Second polynomial.
 * @param n2 Max index of p2, the dimension of p2 is n2+1.
 * @param tmp Temporary array, size at least n1+n2+1.
 */
public void setMult2Pol(final double [] res,final double [] p1,final int n1,final double [] p2,final int n2,final double [] tmp)
{
    for(int k=n1+n2;k>=0;k--) res[k]=0;
    for(int k=n1;k>=0;k--)
    {
	final double c1=p1[k];
	for(int l=n2;l>=0;l--)
	{
	    setQlQmExpansionCoefs(k,l,tmp);
	    TMatr.sumP1(res,tmp,l+k+1,c1*p2[l]);
	}
    }
}




/** Multiply p by ax+b.
 * Calculations in basis.
 * The method should be overloaded if optimal implementation is required.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result, elements=n+2.
 * @param p Input polynomial, elements=n+1.
 */
public void setaxbP(final int n,final double [] res,final double [] p,final double a,final double b)
{
    final double [] f=convertToBasisFromPolynomial_Test(1,new double[]{b,a}),pp=new double [n+1];
    System.arraycopy(p,0,pp,0,n+1);
    setMult2Pol(res,f,f.length-1,pp,n,new double [n+2]);
}


/** Divide  p by (x-x0), returns the remainder.
 * Calculations in basis.
 * The method should be overloaded if optimal implementation is required.
 * @param n The max dimension of p, n_elements=n+1.
 * @param res The result, n_elements=n.
 * @param p Input polynomial, n_elements=n+1.
 * @param x0 To divide on (x-x0)
 * @return The remainder, p=res*(x-x0)+remainder.
 * Note: the res is in basis, but the remainder is a plain value, should NOT be multiplied by Q0.
 */
public double sdiv1(final int n,final double [] res,final double [] p,final double x0)
{
    if(n==0) return convertToPolynomialFromBasis_Test(0,p)[0];
    final double [] f=convertToBasisFromPolynomial_Test(1,new double[]{-x0,1}),r=new double []{0};
    sdiv(res,r,p,n,f,1);
    return convertToPolynomialFromBasis_Test(0,r)[0];
}


/** Divide  p by d, such that p=q*d+r.
 * Calculations in basis.
 * @param q Quotent, 0..np-nd elements are set, dim=np-nd+1.
 * @param r Remainder, 0..nd-1 elements are set, dim=nd.
 * @param p Polynomial to divide.
 * @param np Max index of p, dim=np+1.
 * @param d Divider.
 * @param nd Max index of d, dim=nd+1.
 * Note: in contrast with class Polynomials, where the size of p & d is passed,
 * here we pass max index (1 less than the number of elements).
 */
public void sdiv(final double [] q,final double [] r,final double [] p,final int np,final double [] d,final int nd)
{
    if(nd<0||np<nd) throw new IllegalArgumentException("np="+np+" nd="+nd);
    final int n=np+1;
    final double [] matr=new double [n*n],tmp=new double [n];
    for(int k=np;k>=0;k--)
    {
	if(k>=nd)
	{
	    for(int i=nd,ki=k-nd;i>=0;i--)
	    {
		setQlQmExpansionCoefs(ki,i,tmp);
		final double c=d[i];
		for(int l=ki+i,o=k*n+l;l>=0;l--,o--)
		{
		    matr[o]+=c*tmp[l];
		}
	    }
	}
	else
	{
	    matr[k*n+k]+=1;
	}
    }
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,n,n,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(q,0,n-nd,Double.NaN);
	java.util.Arrays.fill(r,0,nd,Double.NaN);
	return;
    }
    System.arraycopy(p,0,tmp,0,n);
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,n,n,ipvt,tmp,0);
    System.arraycopy(tmp,nd,q,0,n-nd);
    System.arraycopy(tmp,0,r,0,nd);
}


/** Calculates T_n(a(x+x0)) using 3 term recurrence.
 * Calculations in  basis.
 * The b_k^{n} are defined as:
 * T_n(a(x+x0))=sum_{k=0}^{k=n}b_k^{n} T_k(x)
 * The method should be applied consequently to get b_k^{n} for n=0,1,2,3,...
 * @param n The max dimension of b.
 * @param b On input: b_k^{n-1}, on output b_k^{n}.
 * @param q On input: b_k^{n-2}, on output: b_k^{n-1}.
 * The result can be calculated in a variety of ways, e.g.
 * using 3 term recurrence and setaxbP one get
 * b^{n}=(c*x+d)*b^{n-1}+e*b^{n-2}, where c,d,e are the constants,
 * depending on a,x0 and 3 term reccurence coefficients.
 */
public abstract void setNextLinB(final int n,final double [] b,final double [] q,final double a,final double x0);


/** Calculate <F|xTk> moments for k=0..n-1 from <F|Tk>, k=0..n
 * The method should be overloaded if optimal implementation is required.
 *  @param n Max index.
 *  @param d Output, d[k]=<F|xTk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Tk>, k=0..n;
 */
public void setXQnMomentsFromQnMoments(final int n,final double [] d,final double [] moments)
{
    final double [] r=new double [n],mm=new double [n+1];
    for(int k=n;--k>=0;)
    {
	r[k]=1;
	setaxbP(k,mm,r,1,0);
	r[k]=TMatr.mulVV(k+2,mm,moments);
    }
    System.arraycopy(r,0,d,0,n);
}


/** Calculates confederate matrix to solve
 * polynomial equations in basis directly.
 * The leading dimension of returned matrix is coefs.length-1,
 * the number of elements (coefs.length-1)^2.
 * Eigenvalues of confederate matrix are the roots.
 * See:
 * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
 * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
 * and resulting recurrence relations and characterizations", March 2007.
 * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
 * Use the class PolynomialRootsConfederateMatrixABasis
 * to calculate roots for specific basis.
 */
public abstract double [] getConfederateMatrix(final double [] coefs);



} // END of class BasisPolynomials
