package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Laguerre polynomials.
 *  @version $Id: Laguerre.java,v 1.174 2016/03/10 21:10:44 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Laguerre extends BasisPolynomials {


/** Calculates n-th Laguerre polynomial
 * using  L_k=(x*(-1/k)-(1-2*k)/k)*L_{k-1}-((k-1)/k)*L_{k-2}.
 * @param n L index.
 * @param res Output vector.
 * @param Lnm1 Input L_{n-1}
 * @param Lnm2 Input L_{n-2}
 */
protected void setNext(final int nd,final double [] res,
		    final double [] Lnm1,final double [] Lnm2,
		    final boolean flag_type2)
{
    if(nd<0) throw new IllegalArgumentException();
    else if(nd==0)
    {
	res[0]=!flag_type2?1:-1;
    }
    else
    {
	final int n=!flag_type2?nd:nd+1;
	final double nn=n,n1=n-1,nn1=2*n-1;
	res[nd]=-Lnm1[nd-1]/nn;
	if(nd>1) res[nd-1]=(Lnm1[nd-1]*nn1-Lnm1[nd-2])/nn;
	else res[nd-1]=(Lnm1[nd-1]*nn1)/nn;
	for(int k=nd-1;--k>=0;)
	{
	    if(k>0)
		res[k]=(Lnm1[k]*nn1-Lnm1[k-1]-n1*Lnm2[k])/nn;
	    else
		res[k]=(Lnm1[k]*nn1-n1*Lnm2[k])/nn;
	}
    }
}


/**
 * Calculation of the value of the Laguerre polynomial.
 * @param n L index.
 * @param x Argument.
 * @return The value of the Laguerre polynomial Ln at x.
 */
public double calculate(final int n,final double x)
{
    double b=0,s=1;
    for(int i=1;i<=n;++i)
    {
	final double a=b;
	b=s;
	s=(((2*i-1)-x)*b-(i-1)*a)/i;
    }
    return s;
}


public double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [n+1];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double b=0,s=1;
	for(int i=0;i<=n;++i)
	{
	    if(i<=0) m[i]+=s*ww;
	    else {s=((1-i)*b+((2*i-1)-(xx))*(b=s))/i;m[i]+=s*ww;}
	}
    }
    return m;
}

/** Calculates sum_{k=0}^{k=n}L_k(x) Lcoefs_k
 *  using Clenshaw recurrence formula.
 * @param n L index.
 * @param Lcoefs Coefficients of L_n.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public double calculateSumD0(final int n,final double [] Lcoefs,final double x)
{
    double b1=0,b2=0,s=0;
    for(int i=n; i>=0;--i)
    {
	s=((2*i+1)-x)*b1/(i+1)-(i+1)*b2/(i+2)+Lcoefs[i];
	b2=b1;
	b1=s;
    }
    return s;
}


/** Calculates d/dx(sum_{k=0}^{k=n}L_k(x) Lcoefs_k)
 *  using Clenshaw recurrence formula.
 * @param n L index, elements=n+1
 * @param Lcoefs Coefficients of L_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD1(final int n,final double [] Lcoefs,final double x)
{
    double b1=0,b2=0,s=0,sd=0;
    for(int i=n;i>=1;i--)
    {
	s=((2*i-1)-x)*b1/i-i*b2/(i+1)+(sd-=Lcoefs[i]);
	b2=b1;
	b1=s;
    }
    return s;
}

/** Calculates d^2/dx^2(sum_{k=0}^{k=n}L_k(x) Lcoefs_k)
 *  using Clenshaw recurrence formula.
 * @param n L index, elements=n+1
 * @param Lcoefs Coefficients of L_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD2(final int n,final double [] Lcoefs,final double x)
{
    double b1=0,b2=0,s=0,sd=0,sd2=0;
    for(int i=n-1;i>=1;i--)
    {
	s=((2*i-1)-x)*b1/i-i*b2/(i+1)+(sd2-=(sd-=Lcoefs[i+1]));
	b2=b1;
	b1=s;
    }
    return s;
}

/** Differentiate a function in Laguerre basis, formulae
 *  8.971.2, page 1051, and 8.974.3, page 1052, Gradshtein & Ryzhik 1963.
 *  @param n Max index.
 *  @param Lcoefs input, F(x)=L_k(x)*Lcoefs[k]
 *  @return L_k(x)'*Lcoefs[k].
 */
public double [] getDifferentiated(final int n,final double [] Lcoefs)
{
    final double [] d=new double [n];
    double s=0,x=Lcoefs[n];
    for(int k=n;k>=1;)
    {
	s-=x;
	x=Lcoefs[--k];
	d[k]=s;
    }
    return d;
}

/** Integrate a function in Laguerre basis, formulae
 * Prudnikov, Brychkov, Marichev, volume 2, 2003, formulae 1.14.3.1 page 47.
 * @param n Max index.
 * @param Tcoefs Input: F(x)=L_k(x)*Lcoefs[k].
 * @return  \int_0^x F(x)dx=\int_0^x L_k(x)dx*Lcoefs[k].
 */
public double [] getIntegrated(final int n,final double [] Lcoefs)
{
    final double [] i=new double [n+2];
    i[n+1]=-Lcoefs[n];
    for(int k=n;k>=1;--k)
    {
	i[k]=Lcoefs[k]-Lcoefs[k-1];
    }
    i[0]=Lcoefs[0];
    return i;
}

/** Expand Ll(z)Lm(z)=sum_k=0^{l+m} c_k Lk(z),
 *  G.N. Watson London Math. Soc. Jn., v 13, 1938, p.29.
 *  See also: Joseph Gillis and George Weiss, Products of Laguerre Polynomials.
 *  @param l l.
 *  @param m m.
 *  @param c coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] c)
{
    final int r,s;
    if(m<=l) {r=m;s=l;}
    else {s=m;r=l;}
    final int nm=s+r;

    for(int t=0;t<=nm;t++)
    {
	final int p=nm-t,n0=(p+1)/2,ne=Math.min(r,p);
	double cs=0,cfact=0;
	for(int n=n0;n<=ne;n++)
	{
	    if(n<=n0)
	    {
		cfact=1;
		for(int k=nm-n0,ke=r-n0,kse=s-n0,ks=1;k>ke;k--,ks++)
		{
		    cfact*=k;
		    if(ks<=kse) cfact/=ks;
		}
		for(int k=p-n0;k>=1;k--) cfact/=k;
		// System.err.println("p-n0="+(p-n0)+" 2*n0-p="+(2*n0-p)+" p="+p);
		if(p%2!=0) cfact*=-2;
	    }
	    cs+=cfact;
	    //System.err.println("n="+n+" p="+p+" cfact="+cfact);
	    if(false)
	    { // unit test, ft=cfact is always an integer.
		long ft=Polynomials.fact(r+s-n)/
		    Polynomials.fact(r-n)/Polynomials.fact(s-n)/Polynomials.fact(2*n-p)/
		    Polynomials.fact(p-n);
		for(int k=0;k<n;k++) ft*=4;
		for(int k=0;k<p;k++) ft/=(-2);
		UnitTests.checkDoubleRel(ft,"ft",cfact,"cfact{n="+n+" p="+p+"}",1e-7);
	    }
	    if(n<ne)
	    {
		cfact=(cfact*(4L*(r-n)*(s-n)*(p-n)))/(((long)nm-n)*(2*n+1-p)*(2*n+2-p));
	    }
	}
	c[t]=cs;
    }
}



/** Multiply p by ax+b.
 * Calculations in Laguerre.
 * Use Laguerre recurrence, see
 * formulae 8.971.6, MO109, page 1051, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result, elements=n+2.
 * @param p Input polynomial.
 */
@Override
public void setaxbP(final int n,final double [] res,final double [] p,final double a,final double b)
{
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	final double ar=a*r;
	res[k+1]-=(k+1)*(ar);
	res[k]+=(2*k+1)*(ar)+b*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=-k*(ar);
	    r=t;
	}
    }
}


/** Divide  p by (x-x0), returns the remainder.
 * Calculations in Laguerre basis.
 * Use Laguerre recurrence, see
 * formulae 8.971.6, MO109, page 1051, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, n_elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return The remainder.
 */
@Override
public double sdiv1(final int n,final double [] res,final double [] p,final double x0)
{
    double r=p[n];
    //System.err.println("pi="+java.util.Arrays.toString(p));
    if(n>0) res[n-1]=p[n-1];
    for(int k=n;--k>=0;)
    {
	final double c=-r/(k+1),b=(res[k]+=c*(x0-(2*k+1)));
	if(k>0)res[k-1]=p[k-1]+c*k;
	res[k]=c;
	r=b;
    }
    //System.err.println("p="+java.util.Arrays.toString(p));
    //System.err.println("res="+java.util.Arrays.toString(res));
    return r;
}




/** Calculates L_n(a(x+x0)) using 3 term recurrence.
 * Calculations in Laguerre basis.
 * The b_k^{n} are defined as:
 * L_n(a(x+x0))=sum_{k=0}^{k=n}b_k^{n} L_k(x)
 * The method should be applied consequently to get b_k^{n} for n=0,1,2,3,...
 * Uses Laguerre recurrence, see
 * formulae 8.971.6, MO109, page 1051, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of b.
 * @param b On input: b_k^{n-1}, on output b_k^{n}.
 * @param q On input: b_k^{n-2}, on output: b_k^{n-1}.
 */
public void setNextLinB(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    final int n1=n-1;
    double r=b[n1];
    b[n1]=b[n]=0;
    final double fn=(n-1)/((double)n),an=a/n,fn2=1+fn;
    for(int k=n;--k>=0;)
    {
	b[k+1]+=(k+1)*r*an;
	b[k]+=(fn2-(2*k+1+x0)*an)*r;
	if(k<n1) b[k]-=fn*q[k];
	q[k]=r;
	if(k>0)
	{
	    final double bb=b[k-1];
	    b[k-1]=k*r*an;
	    r=bb;
	}
    }
}


/** Calculate <F|xLk> moments for k=0..n-1 from <F|Lk>, k=0..n
 *  @param n Max index.
 *  @param d Output, d[k]=<F|xLk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Lk>, k=0..n;
 */
@Override
public void setXQnMomentsFromQnMoments(final int n,final double [] d,final double [] Pcoefs)
{
    double pk1=0,pk=Pcoefs[0];
    for(int k=0;k<n;++k)
    {
	d[k]=-k*pk1+(2*k+1)*(pk1=pk)-(k+1)*(pk=Pcoefs[k+1]);
    }
}



/** Calculates confederate matrix to solve
 * polynomial equations in Laguerre basis directly.
 * Eigenvalues of confederate matrix are the roots.
 * See:
 * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
 * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
 * and resulting recurrence relations and characterizations", March 2007.
 * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
 * We use the three term recurrence:
 * L_k=(x*(-1/k)-(1-2*k)/k)*L_{k-1}-((k-1)/k)*L_{k-2}.
 * Use the PolynomialRootsConfederateMatrixABasis.ABASISLAGUERRE
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	final int k=m+1;
	matr[m*n+m]=-(1-2*k);         // delta/alfa
	if(m>0) matr[(m-1)*n+m]=1-k;  // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=-k; // 1/alfa
	else
	{
	    final double f=-coefs[n]/k; // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}



private static final Laguerre BASIS=new Laguerre();



/** Multiply p by x.
 * Calculations in Laguerre.
 * Use Laguerre recurrence, see
 * formulae 8.971.6, MO109, page 1051, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
public static void setxP(final int n,final double [] res,final double [] p)
{
    BASIS.setaxbP(n,res,p,1,0);
}


public static double get2LaguerreAveraged(final int n,final double [] p1,final double [] p2)
{
    double s=0;
    for(int k=0;k<=n;k++)
    {
	s+=p1[k]*p2[k];
    }
    return s;
}

public static double [] getL1FromL0(final int n,final double [] l0)
{
    final double [] l1=new double [n+1];
    setL1Moments(n,l1,l0);
    return l1;
}

/** Calculate <F|L^1_n> moments.
 *  @param n Max index.
 *  @param d Output, d[k]=<F|L^1_k>.
 *  @param l0 Input, l0[k]=<F|L^0_k>.
 */
public static void setL1Moments(final int n,final double [] d,final double [] l0)
{
    double s=0;
    for(int k=0;k<=n;++k)
    {
	d[k]=(s+=l0[k]);
    }
}



/** \int_x^{infinity} f(x) exp(-x) dx = i(x) exp(-x).
 * Prudnikov, Brychkov, Marichev, volume 2, 2003,
 * formulae 1.14.3.6 page 47.
 */
public static double [] getLaguerreExpFIntegrated(final int n,final double [] Lcoefs)
{
    final double [] i=new double [n+1];
    //java.util.Arrays.fill(i,0,n+1,0);
    for(int k=n;k>=1;--k)
    {
	final double f=Lcoefs[k];
	i[k]+=f;
	i[k-1]-=f;
    }
    i[0]+=Lcoefs[0];
    return i;
}





/** A different way to calculate roots. Use Jacobi matrix.
 */
static double [] getRoots(final int n,final double x0,final boolean flag_radau)
{
    final double [] a=new double [n+1],b=new double [n];
    a[0]=1;
    for(int k=0;k<n;k++)
    {
	b[k]=2*k+1;
	a[k+1]=k+1;
    }
    if(flag_radau)
    {
	// multiply by -1 because Laguerre polynomials have a sign (-1)^n
	// before highest monomial.
	b[n-1]+=(-1)*a[n]*(BASIS.calculate(n,x0)/BASIS.calculate(n-1,x0));
    }
    return EVSolver.getEV_3DCommon(n,b,0,a,1,false);
}
public static double [] getRoots(final int n)
{
    return getRoots(n,Double.NaN,false);
}
public static double [] getRadauRoots(final int n)
{
    return getRoots(n,0,true);
}
/** Uses Abramowitz and Stegun 1972, p. 890.
 */
public static double [] getWeights(final double [] xx)
{
    final int n=xx.length;
    final double [] w=new double [n];
    final double f=1.0/((n+1)*(n+1));
    for(int k=n;--k>=0;)
    {
	final double x=xx[k];
	w[k]=f*x/Math.pow(BASIS.calculate(n+1,x),2);
    }
    return w;
}
public static double [] getNodesPolynomial(final int n)
{
    final double [] nodes=new double [n+1];
    nodes[n]=1;
    return nodes;
}
public static double [] getRadauNodesPolynomial(final int n)
{
    final double [] nodes=new double [n+1];
    nodes[0]=1;
    setL1Moments(n-1,nodes,nodes); // get delta function by differentiating a step function
    setxP(n-1,nodes,nodes);
    return nodes;
}

/** Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 * Mathematics and Computers in Simulation 54 (2000) 403-412.
 * https://www.cs.purdue.edu/homes/wxg/selected_works/section_08/164.pdf
 */
public static double [] getRadauWeights(final double [] xr)
{
    final int n=xr.length;
    final double [] w=new double [n];
    final double f=1.0/n;
    for(int k=n;--k>=0;)
    {
	w[k]=f/Math.pow(BASIS.calculate(n-1,xr[k]),2); // 5.9
    }
    return w;
}
/** Christoffel function. */
public static double [] getChristoffelKernel(final int n)
{
    final double [] c=new double [2*n+1],tmp=new double [2*n+1];
    for(int k=0;k<=n;k++)
    {
	BASIS.setQlQmExpansionCoefs(k,k,tmp);
	TMatr.sumP1(c,tmp,2*k+1,1);
    }
    return c;
}






/** The same as setQlQmExpansionCoefs,
 *  but uses 3 term recursion instead of hypergeometric function expansion.
 *  It is rather slow at large l,m. Used for unit test.
 */
public static void setQlQmExpansionCoefs_test(final int l,final int m,final double [] s)
{
    for(int k=l+m;k>=0;k--) s[k]=0;
    final int ll,mm;
    if(l<=m) {ll=l;mm=m;}
    else {ll=m;mm=l;}
    sumQlQmExpansionCoefs(ll,mm,s,1);
}
private static void sumQlQmExpansionCoefs(final int l,final int m,final double [] s,final double c)
{
    if(l>m) throw new IllegalArgumentException("l="+l+">m="+m);
    if(l<=0) {s[m]+=c; return;}
    if(l>=2) sumQlQmExpansionCoefs(l-2,m,s,c*(-1+1.0/l));
    sumQlQmExpansionCoefs(l-1,m-1,s,c*m/l);
    sumQlQmExpansionCoefs(l-1,m,s,c*(2-(2*m+2)/((double)l)));
    sumQlQmExpansionCoefs(l-1,m+1,s,c*(m+1)/l);
}



/** A class for unit test. */
static class TestBasisLaguerre extends TestBasisPolynomials {

TestBasisLaguerre()
{
    this(new java.util.Random(-790811),1e-6);
}


TestBasisLaguerre(final java.util.Random r,final double eps)
{
    super("Laguerre",BASIS,r,eps);
}

double [] getMonomialsMoments(final int n)
{
    final double [] pmom=new double [n];
    for(int k=0;k<pmom.length;k++) pmom[k]=Polynomials.fact(k); // monomials moments
    return pmom;
}

OrthogonalPolynomialsABasis getOrtPol()
{
    return new OrthogonalPolynomialsLaguerreBasis();
}

double [][] getExact()
{
    return new double [][]{
	{1},
	{1,-1},
	TMatr.getVectorScaled(new double []{2,-4,1},1.0/2),
	TMatr.getVectorScaled(new double []{6,-18,9,-1},1.0/6),
	TMatr.getVectorScaled(new double []{24,-96,72,-16,1},1.0/24),
	TMatr.getVectorScaled(new double []{120,-600,600,-200,25,-1},1.0/120),
	TMatr.getVectorScaled(new double []{720,-4320,5400,-2400,450,-36,1},1.0/720),
	TMatr.getVectorScaled(new double []{5040,-35280,52920,-29400,7350,-882,49,-1},1.0/5040),
	TMatr.getVectorScaled(new double []{40320,-322560,564480,-376320,117600,-18816,1568,-64,1},1.0/40320),
	TMatr.getVectorScaled(new double []{362880,-3265920,6531840,-5080320,1905120,-381024,42336,-2592,81,-1},1.0/362880)
    };
}

private void printAllQuadratures(final int nmax)
{
    final PolynomialRootsFinder rf=new PolynomialRootsFinder(){
	    public int findRoots(final int n,final double [] p,final double [] re_root,final double [] im_root) {
		return new PolynomialRootJenkinsTraubORIG(n).findRoots(n,BASIS.convertToPolynomialFromBasis_Test(n,p),re_root,im_root);
	    }};
    for(int n=1;n<=nmax;n++)
    {
	printTestRoots(rf,n);
	printQuadratures(rf,n);
	printRadauQuadratures2(rf,n,r);
	printDeltaFunctionRoots(rf,n);
    }
}


void printQuadratures(final PolynomialRootsFinder rf,final int n)
{
    System.err.println("=========Laguerre: Gauss quadratures for n="+n);
    final double eps=1e-4;

    final double [] pol=new double [n+1];pol[n]=1;
    final double [] root_re=new double [n],root_im=new double [n];
    final int nroots=rf.findRoots(n,pol,root_re,root_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoad(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    final double [] wh=getWeights(xr),ww=new double [xr.length];
    final double [] kk=getChristoffelKernel(n),wk=new double [ww.length];

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double w=-1/(n*BASIS.calculate(n-1,root_re[k])*BASIS.calculateSumD1(n,pol,root_re[k]));
	s+=w;
	S-=w*Math.log(w);
	ww[k]=w;
	wk[k]=1.0/BASIS.calculateSumD0(kk.length-1,kk,root_re[k]);
	System.err.println(String.valueOf(root_re[k])+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    UnitTests.checkDoubleArrayRel(ww,"ww",wh,"wh",eps);
    if(n<15)
    UnitTests.checkDoubleArrayRel(wk,"wk",wh,"wh",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();
}
void printTestRoots(final PolynomialRootsFinder rf,final int n)
{
    final double eps=1e-5;
    final double [] pol=new double [n+1];pol[n]=1;
    final double [] root_re=new double [n],root_im=new double [n];
    final int nroots=rf.findRoots(n,pol,root_re,root_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoad(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    System.err.println("=========Laguerre: TestRoots for n="+n);
    System.err.println("Xm\tFm\tXf\tFf");
    for(int k=0;k<n;k++)
    {
	System.err.println(String.valueOf(xr[k])+"\t"+BASIS.calculate(n,xr[k])+"\t"+
			   root_re[k]+"\t"+BASIS.calculate(n,root_re[k]));
    }
    System.err.println();
}

void printRadauQuadratures2(final PolynomialRootsFinder rf,final int neq,final java.util.Random r)
{
    System.err.println("=========Laguerre-Radau: Gauss quadratures for neq="+neq);
    final double eps=1e-4;

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] matr=new double [neq*neq],
	pol=new double [neq+1],e=new double [2*neq+1];
    for(int k=0;k<neq;k++)
    {
	BASIS.setQlQmExpansionCoefs(neq,k,e);
	setxP(neq+k,e,e);
	//System.err.println("e="+java.util.Arrays.toString(e));
	pol[k]=-TMatr.mulVV(neq+k+2,mom,e);
	for(int l=0;l<=k;l++)
	{
	    BASIS.setQlQmExpansionCoefs(l,k,e);
	    setxP(l+k,e,e);
	    matr[k+l*neq]=matr[l+k*neq]=TMatr.mulVV(l+k+2,e,mom);
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0) throw new RuntimeException("pinfo="+pinfo);
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,pol,1);
    pol[neq]=1;
    //System.err.println("pol="+java.util.Arrays.toString(pol));
    final double [] xs=new double [neq+1],root_im=new double [neq+1];
    final int nroots=rf.findRoots(neq,pol,xs,root_im);
    if(nroots!=neq) throw new RuntimeException("nroots="+nroots+" neq="+neq);
    xs[neq]=0;root_im[neq]=0; // radau node
    final double [] rpol=new double [neq+2],tmp=new double [neq+1];
    setxP(neq,rpol,pol);
    if(true)
    {
	final double [] nrr=getRadauNodesPolynomial(rpol.length-1);
	for(int k=0;k<rpol.length;k++) nrr[k]*=rpol[rpol.length-1]/nrr[rpol.length-1];
	UnitTests.checkDoubleArrayRel(nrr,"nrr",rpol,"rpol",eps);
    }
    final double [] ws=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	if(Double.compare(x,0.0)==0)
	{ // radau node
	    ws[k]=TMatr.mulVV(neq+1,pol,mom)/BASIS.calculateSumD0(neq,pol,x);
	}
	else
	{
	    BASIS.sdiv1(neq+1,tmp,rpol,x);
	    ws[k]=TMatr.mulVV(neq+1,tmp,mom)/BASIS.calculateSumD0(neq,tmp,x);
	}
    }
    SortArrays.sortArrayWithLoad(neq+1,xs,ws);

    final double [] xr=getRadauRoots(neq+1);
    SortArrays.sortArrayWithLoad(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",xs,"xs",eps);

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	S-=w*Math.log(w);
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();

    // test radau exact integral, polynomial of 2*neq power
    if(neq<12)
    for(int t=0;t<1000;t++)
    {
    final double [] pt=new double [2*neq +1];
    for(int k=pt.length;--k>=0;) pt[k]=r.nextDouble();
    pt[0]+=0.6+r.nextDouble()+r.nextDouble();
    double i=0;
    for(int k=0;k<xs.length;k++)
    {
	i+=ws[k]*BASIS.calculateSumD0(pt.length-1,pt,xs[k]);
    }
    final double ia=TMatr.mulVV(pt.length,pt,mom);
    UnitTests.checkDoubleRel(i,"i",ia,"ia",eps);
    }

    // check Radau exact weights.
    if(neq<=7)
    checkRadauExactWeights(xs,ws,eps);
}

/** Check Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 */
void checkRadauExactWeights(final double [] xs,final double [] ws,final double eps)
{
    final int neq=xs.length-1;
    UnitTests.checkDoubleRel(ws[0],"ws[0]",1.0/(neq+1),"1.0/(neq+1)",eps);  // 6.5
    UnitTests.checkDoubleArrayRel(ws,"ws",getRadauWeights(xs),"getRadauWeights(xs)",eps);
    System.err.println("checkRadauExactWeights: Radau exact weights OK, neq="+neq);
}

void printDeltaFunctionRoots(final PolynomialRootsFinder rf,final int neq)
{
    System.err.println("=========Laguerre: Quadratures on (x)delta(x) roots for neq="+neq);
    final double eps=1e-5;
    final double [] o=new double [neq+1];
    o[0]=1;
    setL1Moments(neq,o,o); // get delta function by differentiating a step function

    //System.err.println("o="+java.util.Arrays.toString(o));

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] wnodes=new double [neq+2],tmp=new double [neq+2];
    BASIS.setMult2Pol(wnodes,o,o.length-1,new double []{1,-1},1,tmp);
    final double [] xs=new double [neq+1],root_im=new double [neq+1];
    final int nroots=rf.findRoots(neq+1,wnodes,xs,root_im);
    if(nroots!=neq+1) throw new RuntimeException("nroots="+nroots+" neq+1="+(neq+1));
    final double [] ws=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	BASIS.sdiv1(neq+1,tmp,wnodes,x);
	ws[k]=TMatr.mulVV(neq+1,tmp,mom)/BASIS.calculateSumD0(neq,tmp,x);
    }

    SortArrays.sortArrayWithLoad(neq+1,xs,ws);
    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println();
    if(neq<=7)
    checkRadauExactWeights(xs,ws,eps);
}


void test_setQlQmExpansionCoefs(final int n_tests)
{
    super.test_setQlQmExpansionCoefs(n_tests);

    for(int t=0;t<n_tests;t++)
    {

    for(int l=0;l<=nmax;l++)
    for(int m=0;m<=nmax;m++)
    {
	final double [] res=new double [l+m+1],resa=new double [l+m+1];
	for(int k=0;k<res.length;k++){res[k]=r.nextDouble();resa[k]=r.nextDouble();}
	BASIS.setQlQmExpansionCoefs(l,m,res);
	setQlQmExpansionCoefs_test(l,m,resa);
	UnitTests.checkDoubleArrayRel(res,"res",resa,"resa",eps);
	//System.err.println("l="+l+" m="+m+" test OK");
    }
    }
    System.err.println("Laguerre:test_setQlQmExpansionCoefs(): EXTRA test OK with nmax="+nmax+" eps="+eps);
}


private void test_getLaguerreExpFIntegrated(final java.util.Random r,final int n_tests,final int n_max,final double eps)
{

for(int n=0;n<=n_max;n++)
{
    final double [] p=new double [n+1];
    for(int t=0;t<n_tests;t++)
    {
	for(int i=p.length;--i>=0;) p[i]=r.nextDouble();

	final double [] ip=getLaguerreExpFIntegrated(p.length-1,p);
	final double [] mp=BASIS.convertToPolynomialFromBasis_Test(p.length-1,p);
	for(int k=0;k<mp.length;k++) if(0!=k%2) mp[k]=-mp[k];
	// Prudnikov, vol 1, page 113, formulae 1.3.2.9.
	final double [] s=new double [mp.length];
	double [] d=mp;
	for(int k=0,sign=1;k<mp.length;k++,sign=-sign)
	{
	    TMatr.sumP1(s,d,sign);
	    d=Polynomials.getPolynomialDifferentiated(d);
	}
	for(int k=0;k<s.length;k++) if(0!=k%2) s[k]=-s[k];
	UnitTests.checkDoubleArrayRel1(BASIS.convertToPolynomialFromBasis_Test(n,ip),"conv(ip)",s,"s",eps);
	UnitTests.checkDoubleArrayRel1(BASIS.convertToBasisFromPolynomial_Test(n,s),"conv(s)",ip,"ip",eps);

	final double [] rp=BASIS.convertToPolynomialFromBasis_Test(p.length-1,p);
	UnitTests.checkDoubleArrayRel1(BASIS.convertToPolynomialFromBasis_Test(n,ip),"conv(ip)",Polynomials.getPolynomialExpFIntegrated(rp.length,rp),"I(rp)",eps);
	UnitTests.checkDoubleArrayRel1(s,"s",Polynomials.getPolynomialExpFIntegrated(rp.length,rp),"I(mp)",eps);
    }
}

System.err.println("Laguerre:test_getLaguerreExpFIntegrated(): test OK with n_max="+n_max+" n_tests="+n_tests+" eps="+eps);
}


private void test_LnprimeMinusLn(final java.util.Random r,final int n_max,final int n_tests,final double eps)
{
    for(int t=0;t<n_tests;t++)
    {
	final double [] pol=new double [1+r.nextInt(1+n_max)];
	final int n=pol.length-1;
	for(int k=n;k>=0;--k){pol[k]=r.nextDouble();}

	if(true) // Test setL1Moments.
	{
	    final double [] d=new double [n+1];
	    setL1Moments(n,d,pol);

	    if(true) // Test against direct calculation.
	    {
		final double [] xd=new double [n+1];
		double [] ra=new double [n+1],rb=new double [n+1];
		for(int k=0;k<=n;k++)
		{
		    BASIS.setNext(k,rb,ra,rb);
		    final double [] l=rb;rb=ra;ra=l;
		    final double [] xx=new double [k+1];
		    xx[k]=l[k];for(int j=0;j<k;j++) {xx[j]=l[j]-(j+1)*l[j+1];} // L-L'
		    final double [] xxl=BASIS.convertToBasisFromPolynomial_Test(k,xx); // convert back to Laguerre
		    xd[k]=get2LaguerreAveraged(k,xxl,pol);
		}
		UnitTests.checkDoubleArrayRel(d,"d",xd,"xd",eps);
	    }
	}
	if(true)
	{
	    final double [] Xpp=new double [n];
	    for(int k=0;k<n;k++)
	    {
		final double [] tmp=new double [k+1],rtmp=new double [k+2];
		tmp[k]=1;
		setxP(k,rtmp,tmp);
		Xpp[k]=TMatr.mulVV(k+2,rtmp,pol);
	    }
	    final double [] sXpp=new double [n];
	    BASIS.setXQnMomentsFromQnMoments(n,sXpp,pol);
	    UnitTests.checkDoubleArrayRel(sXpp,"sXpp",Xpp,"Xpp",eps);
	    final double [] sXpp_same=pol.clone();
	    BASIS.setXQnMomentsFromQnMoments(n,sXpp_same,sXpp_same);
	    UnitTests.checkDoubleArrayRelFirstindex(n,sXpp_same,"sXpp_same",Xpp,"Xpp",eps);
	}
    }
    System.err.println("Laguerre:test_LnprimeMinusLn() test OK with eps="+eps);
}


void printExtra(final int nm)
{
    printAllQuadratures(25);
}
void test_Extra(final int n_tests)
{
    test_LnprimeMinusLn(r,nmax>10?10:nmax,n_tests,eps);
    test_getLaguerreExpFIntegrated(r,n_tests,nmax,eps);
}

} // END OF class TestBasisLaguerre


public static void main(String [] args)
{
    new TestBasisLaguerre().runTests("$Id: Laguerre.java,v 1.174 2016/03/10 21:10:44 mal Exp $");
}

} // END of class Laguerre
