package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Manupulate orthogonal polynomials in arbitrary basis.
 *  @version $Id: OrthogonalPolynomialsABasis.java,v 1.192 2015/10/23 11:41:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class OrthogonalPolynomialsABasis {

/** Not always available! Many be null if basis operations directly overloaded.
 */
public final BasisPolynomials B;
public OrthogonalPolynomialsABasis(final BasisPolynomials B){this.B=B;}

public EVSolver getEVSolver(){return new EVSolver();}
public final EVSolver EV=getEVSolver();


/** Calculates <Q|Q> matrices.
 * @param n Matrix dimension.
 * @param mom Moments, at least 2n-1 elements should present (0..2n-2 moments).
 */
public double [] getQQMatr(final int n,final double [] mom)
{
    final double [] matr=new double [n*n],e=n<=0?null:new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<=k;l++)
	{
	    setQQCoefs(l,k,e);
	    matr[k+l*n]=matr[l+k*n]=TMatr.mulVV(l+k+1,e,mom);
	}
    }
    return matr;
}
/** Calculates Q(x)<QQ>Q(x) polynomial.
 *  @param n The dimension to use.
 *  @param QQ The <QQ> matrix.
 *  @param nm The leading dimension of QQ.
 */
public double [] getKK(final int n,final double [] QQ,final int nm)
{
    final double [] r=new double [2*n-1];
    final double [] tmp=new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    setQQCoefs(l,k,tmp);
	    TMatr.sumP1(r,tmp,1+k+l,QQ[k*nm+l]);
	}
    }
    // Q(x)<QQ> Q(x) is now formed
    return r;
}


/** Calculate first n orhogonal polynomials (0..n-1)
 * from first 2n-1 moments.
 * All we need is <Q_l Q_m> matrix which is internally
 * constructed from mom[k]=<Q_k> by getQQMatr() using setQQCoefs().
 * @param n The number of polynomials to generate.
 * @param mom Moments, must be at least 2n-1 elements (0..2n-2 moments).
 * @return Orthogonal Polynomials.
 */
public double [][] getOrthogonalPolynomialsFirstKind(final int n, final double [] mom)
{
    final double [][] p=getOrthogonalized(n,getQQMatr(n,mom),n);
    for(int k=0;k<p.length;k++)
    {
	final double [] pk=p[k];
	final double sg=getSignToMatch(k);
	for(int j=pk.length;--j>=0;) pk[j]=sg*pk[j];
    }
    return p;
}

/** Calculates single orthogonal polynomial.
 *  @param n Order of the polynomial.
 *  @param moms Moments in basis, should have at least 2*n elements in array (0..2n-1 moments).
 *  @return Polynomial, normalized as P[n]=1.
 *  Used for testing.
 */
public double [] getSingleOrthogonalPolynomial(final int n,final double [] moms)
{
    final double [] matr=getQQMatr(n,moms),P=new double [n+1],tmp=new double [2*n];
    for(int k=0;k<n;k++)
    {
	setQQCoefs(n,k,tmp);
	P[k]=-TMatr.mulVV(1+n+k,moms,tmp);
    }
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,n,n,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(P,Double.NaN);
	return P;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,n,n,ipvt,P,1);
    P[n]=1;
    return P;
}

/** The same as getGrammSchmidtOrthogonalized, slower, but hopefully? more stable. The method is equivalent to repeatedly calling getSingleOrthogonalPolynomial with subsequent normalizing.
 * Calculations can be performed in various basises Q_k,
 * @param n The number of basis vectors to build.
 * @param QQ The <Q_i Q_j> matrix, i,j=0..n-1.
 * @param nm The leading dimension of QQ.
 */
public static double [][] getOrthogonalized(final int n,final double [] QQ,final int nm)
{
    final double [] e=new double [nm*n];
    final int [] ipvt=new int [n-1];
    // it may be more stable to solve linear system all at once several times
    // rather than solving it incrementally
    final double [][] p=new double [n][];
    for(int k=0;k<n;k++)
    {
	System.arraycopy(QQ,0,e,0,k*nm);
	final double [] pk=p[k]=new double [k+1];
	pk[k]=1;
	for(int l=0;l<k;l++) pk[l]=-QQ[k*nm+l];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(e,k,nm,ipvt);
	if(pinfo!=0) java.util.Arrays.fill(pk,Double.NaN);
	else com.polytechnik.utils.Linsystems.dgesl_m1(e,k,nm,ipvt,pk,1);
	// normalize
	final double anorm=Math.sqrt(TMatr.getAMatrA(k+1,pk,QQ,nm));
	for(int l=0;l<=k;l++) pk[l]/=anorm;
    }
    return p;
}
/** Orthogonal polynomials are defined within a factor of +-1 multiplied by.
 *  This method return this factor to match standard definition,
 *  corresponding positive sign of the coefficient with maximal x^n,
 *  when alfa_n Q_n(x) is converted to monomials.
 * @return +1 or -1.
 */
public double getSignToMatch(final int k){return 1;}



/** Calculate a & b recurrence coefficients from a Gauss quadrature.
 *  Used for unit test.
 */
public double [][] getABFromQuadrature(final double [] x,final double [] w)
{
    return getAB(getOrthogonalPolynomialsFirstKind(x.length,getMomentsFromQuadrature(x,w)));
}

/** Calculate a & b recurrence coefficients.
 * The definitions as in
 * Walter Gautschi, "Orthogonal Polynomials:Computation and approximation", page 12.
 * Here a[k]=sqrt(beta_k) and b[k]=alfa_k.
 * And the recurrence is: xp_{k}=a_{k+1}*p_{k+1}+b_{k}p_{k}+a_{k}p_{k-1}.
 */
public double [][] getAB(final double [][] p)
{
    final int n=p.length,ne=n-1;
    if(n<=0) return new double [][]{{},{}};
    final double [] a=new double [n],b=new double [ne];
    for(int k=ne;--k>=0;)
    {
	final double [] pk=p[k],pk1=p[k+1];
	setxP(a,pk);
	b[k]=(a[k]-(a[k+1]/=pk1[k+1])*pk1[k])/pk[k];
	//System.err.println("k="+k+" a[k+1]="+a[k+1]+" b[k]="+b[k]);
    }
    a[0]=1/fx(p[0],0);
    return new double [][]{a,b};
}

/** Modify b[n-2] to have Jacobi matrix
 * generate (n-1)-th Radau polynomial exactly.
 * See "Journal of Computational and applied Mathematics",
 * Volume 127, Numbers 1|2, 15 January 2001,
 * D.P. Laurie, "Computation of Gauss-type quadrature formulas", p.201-217.
 */
public double [][] getABRadauModified(final int n,final double [] aa,final double [] bb)
{
    if(n>aa.length) throw new RuntimeException("n="+n+" aa.length="+aa.length);
    if(n<=0) return new double [][]{{},{}};
    else if(n<=1) return new double [][]{{aa[0]},{}};
    final double [] a=new double [n],b=new double [n-1];
    System.arraycopy(aa,0,a,0,n);
    System.arraycopy(bb,0,b,0,n-1);

    final double x0=getRadauRoot();

    //System.err.println("getABRadauModified:n="+n+" a.length="+a.length+" b.length="+b.length+" x0="+x0);
    b[n-2]+=a[n-1]*getQxFromAB(n-1,a,b,x0)/getQxFromAB(n-2,a,b,x0);
    return new double [][]{a,b};
}


/** Calculate Q(x) using three term recurence.
 */
public double getQxFromAB(final int n,final double [] a,final double [] b,final double x)
{
    if(n>=a.length) throw new RuntimeException("n="+n+" a.length="+a.length);
    double f=1/a[0],fprev=0;
    for(int k=0;k<n;k++)
    {
	f=(-a[k]*(fprev)+(x-b[k])*(fprev=f))/a[k+1];
    }
    return f;
}
/** Calculate psi_jQ_j(x) using three term recurence and Clenshaw recurrence formula.
 * See L. Fox, I. B. Parker (1968), Chebyshev Polynomials in Numerical Analysis, page 56.
 */
public double getSUMQxFromAB(final double [] psi,final double x,final double [] a,final double [] b)
{
    /*
    // replace by Clenshaw recurrence formula.
    double s=0;
    for(int k=psi.length;--k>=0;)
    {
	s+=psi[k]*getQxFromAB(k,a,b,x);
    }
    */

    final int n=psi.length-1;
    if(n>=a.length) throw new RuntimeException("n="+n+" a.length="+a.length);
    if(n<0) return 0;
    /*
    double f=0,fprev=0,s=0;
    for(int k=-1;k<n;)
    {
	if(k<0) {f=1/a[0];fprev=0;}
	else {f=(-a[k]*(fprev)+(x-b[k])*(fprev=f))/a[k+1];}
	s+=f*psi[++k];
    }
    return s;
    */
    double sprev=0,s=0;
    for(int i=n,n1=n-1; i>=0;--i)
    {
	s=sprev*((i>=n1)?0:-a[i+1]/a[i+2])+(sprev=s)*(i>=n?0:(x-b[i])/a[i+1])+psi[i];
    }
    // Take into account that Q_{-1}=0 and this lead to
    // second factor in formulae (44) (L. Fox, I. B. Parker, p.56) being equal to 0.
    return s/a[0];
}

/** Build orthogonal polynomials from a & b recurrence coefficients.
 *  May not be stable, but useful for second kind polynomials generation.
 */
public double [][] getOrthogonalPolynomialsFromAB(final int kind,final double [] a,final double [] b)
{
    if(kind>=a.length) return new double [][]{};
    // generate p - kind-type polynomials.
    final double [][] p=new double [a.length-kind][];
    p[0]=new double []{1};p[0][0]/=(fx(p[0],0)*a[kind]);
    for(int k=1;k<p.length;k++)
    {
	final int km1=k-1;
	final double [] pk=p[k]=new double [k+1],pk1=p[km1],pk2=k>1?p[k-2]:null;
	// save p_{k} to p[k];
	setxP(pk,pk1);
	final double ak=a[k+kind],bk1=b[km1+kind],ak1=a[km1+kind];
	pk[k]/=ak;
	pk[km1]=(pk[km1]-bk1*pk1[km1])/ak;
	for(int l=km1;--l>=0;) pk[l]=(pk[l]-bk1*pk1[l]-ak1*pk2[l])/ak;
    }
    return p;
}

public double [][] getSecondKindOrthogonalPolynomialsFromAB(final double [] a,final double [] b)
{
    return getOrthogonalPolynomialsFromAB(1,a,b);
}
public double [][] getOrthogonalPolynomialsFirstKindFromAB(final double [] a,final double [] b)
{
    return getOrthogonalPolynomialsFromAB(0,a,b);
}
public double [][] getSecondKindOrthogonalPolynomials(final double [][] p)
{
    final double [][] ab=getAB(p);
    return getSecondKindOrthogonalPolynomialsFromAB(ab[0],ab[1]);
}

/** Recover 0 .. 2n-1 moments from 0 .. n  orthogonal polynomials.
 *  The 2n-th moment is the normalizing factor of p[n] and cannot be recovered.
 *  Used for unit tests. For production use getMomentsFrom_AB.
 */
public double [] getMomentsFromPolynomial(final double [][] p)
{
    final int n=p.length-1;
    if(n<=0) throw new RuntimeException("p.length="+p.length+"<=1");
    final double [] xs=getXNodes(p[n]),ws=getWeightsChristoffel(xs,p);
    return getMomentsFromQuadrature(xs,ws);
}


/** Recover 0 .. 2n-1 moments from a[0..n] b[0..n-1].
 *  Used for unit tests and to recover moments for second kind
 *  orthogonal polynomials measure.
 */
public double [] getMomentsFrom_AB(final double [] a, final double [] b)
{
    final int n=a.length-1;
    if(n<=0) throw new RuntimeException("a.length="+a.length+"<=1");
    final double [][] nd=getQuadraturesFromAB_EV(n,a,b);
    return getMomentsFromQuadrature(nd[0],nd[1]);
}
/** Given Gauss quadrature(x&w) recover 2n moments (0..2n-1 moments).
 */
public double [] getMomentsFromQuadrature(final double [] x,final double [] w)
{
    final int n=x.length;
    final double [] mom=new double [2*n];
    // Use Gauss quadrature integration.
    for(int k=mom.length;--k>=0;)
    {
        final double [] t=new double [k+1];t[k]=1;
        // integrate
        double s=0;
        for(int l=0;l<x.length;l++) s+=w[l]*fx(t,x[l]);
        mom[k]=s;
    }
    return mom;
}


/** Weights by integration.
 *  Integration is replaced by synthetic division by basis[] polynomials.
 *  The coefficient by basis[0] give the value of integral.
 *  The weights are normalized as sum(weights)=<1>.
 */
public double [] getWeights(final double [] nodes,final double [] x,final double [][] basis)
{
    final double [] w=new double [x.length],tmp=new double [nodes.length-1];
    final double f=fx(basis[0],0);
    for(int k=x.length;--k>=0;)
    {
	sdiv1(tmp,nodes,x[k]);
	final double t0=fx(tmp,x[k]);
	TMatr.expandOverLowerDiagBasis(tmp,tmp,tmp.length,basis);
	w[k]=(tmp[0]/t0)/f;
    }
    return w;
}
/** Weights using Christoffel function.
 *  Here x are the roots of p[x.length], and
 *  the weights are normalized as sum(weights)=<1>.
 */
public double [] getWeightsChristoffel(final double [] x,final double [][] p)
{
    final int n=x.length;
    if(n<=0) return new double []{};
    else if(n<=1) return new double []{1/Math.pow(fx(p[0],0),2)};
    final double [] nodes=p[n],pp=p[n-1],ws=new double [n];
    sdiv1(ws,nodes,0);
    final double f=ws[n-1]/pp[n-1];
    for(int k=0;k<n;k++)
    {
	final double xx=x[k];
	ws[k]=f/(fx(pp,xx)*Dfx(nodes,xx));
    }
    return ws;
}
/** Weights by integration.
 *  Here the x[] are the roots of nodes polynomial.
 *  Synthetically divide nodes by (x-x[k]), then integrate Legendre interpolatory polynomial directly using known values of moms[k]=<Q_k(x)>.
 *  Typically stable (except for exponentially small weights).
 *  The weights are normalized as sum(weights)=<1>.
 * @param nodes Nodes polynomial.
 * @param x Points.
 * @param moms Moments, at least nodes.length-1 elements must present.
 * @return The weights.
 */
public double [] getWeightsByIntegration(final double [] nodes,final double [] x,final double [] moms)
{
    final double [] w=new double [x.length];
    final int n=nodes.length-1;
    final double [] tmp=new double [n];
    for(int k=x.length;--k>=0;)
    {
	sdiv1(tmp,nodes,x[k]);
	w[k]=TMatr.mulVV(n,tmp,moms)/fx(tmp,x[k]);
    }
    return w;
}

/** Find roots by finding the eigenvalues of Jacobi matrix, n<a.length.
 */
public double [] getXNodesFromAB(final int n,final double [] a,final double [] b)
{
    if(n>=a.length) throw new IllegalArgumentException("n="+n+">= a.length="+a.length);
    //System.err.println("a="+java.util.Arrays.toString(a)+" b="+java.util.Arrays.toString(b)+" nodes="+java.util.Arrays.toString(nodes));
    return EV.getEV_Thridiagonal(n,b,0,a,1,false);
}

/** Find roots by solving polynomial equation.
 *  Imaginary part is discarded, so double roots possible if nodes polynomial is not well defined.
 *  Use for unit test only.
 */
public double [] getXNodes(final double [] nodespolynomial)
{
    final double [] re=new double [nodespolynomial.length-1],im=new double [nodespolynomial.length-1];
    final int nr=getPolynomialRootsFinderInBasis().findRoots(re.length,nodespolynomial,re,im);
    if(nr!=re.length) {java.util.Arrays.fill(re,Double.NaN);return re;}
    SortArrays.sortArrayWithLoad(re.length,re,im);
    return re;
}


/** The returned [][] array contains 3 [] arrays: x nodes, weights and nodes polynomial.
 *  Corresponding to quadratures built on p[n] polynomial.
 *  Weights are normalized as sum(weights)=<1>;
 *  Used for unit tests.
 */
public double [][] getQuadraturesFromPolynomials(final int n,final double [][] p)
{
    if(n>=p.length) throw new RuntimeException("n="+n+">= p.length="+p.length);
    final double [] nodes=p[n];
    final double [] xs=getXNodes(nodes),ws=getWeightsChristoffel(xs,p);
    return new double [][] {xs,ws,nodes.clone()};
}


public double [][] getRadauPolynomials(final double [][] p)
{
    final int n=p.length-1;
    if(n<=0) return new double [][]{};
    final double [] am=new double [n],cm=new double [n],bm=new double [n-1];
    for(int k=0;k<n;k++)
    {
	final double [] rmp=getRadauMultiplied(p[k]);
	final double a=rmp[k+1]/p[k+1][k+1],b=(rmp[k]-a*p[k+1][k])/p[k][k];
	cm[k]=b;               // <Q_k Q_k *radau>
	if(k<n-1) bm[k]=a;     // <Q_k Q_{k+1} *radau>
	if(k>0) am[k]=bm[k-1]; // <Q_k Q_{k-1} *radau>
    }
    am[0]=Double.NaN;

    final int [] in1=new int [n-1],in2=new int [n-1];
    final double [][] qq=new double [n][],rr=qq;
    final double [] tmp=new double [n-1];
    for(int k=0;k<n;k++)
    {
	final double [] qk=qq[k]=new double [k+1];
	qk[k]=1;
	if(k>0) qk[k-1]=-am[k]; // <Q_k Q_{k-1} *radau>
	com.polytechnik.utils.Progonka3.eliminate(k,am,cm,bm,qk,qk,tmp,in1,in2);
	// normalize
	double s=0;
	for(int l=0;l<=k;l++) {s+=qk[l]*(qk[l]*cm[l]+(l>0?qk[l-1]*am[l]:0)+(l<k?qk[l+1]*bm[l]:0));}
	//System.err.println("sum s="+s);
	s=Math.sqrt(s);for(int l=0;l<=k;l++) qk[l]/=s;
	//System.err.println("s="+s);

	// Multiply matrices. Because they are lower diagonal, we can do it in place.
	final double [] rk=rr[k];
	for(int l=0;l<=k;l++)
	{
	    final double [] bl=p[l];final double qkl=qk[l];
	    for(int m=0;m<l;m++) rk[m]+=qkl*bl[m];
	    rk[l]=qkl*bl[l];
	}
    }
    return rr;
}

/** The returned [][] array contains 3 [] arrays: x nodes, weights and nodes polynomial.
 *  Corresponding to quadratures built on Radau polynomial, constructed from p (measue: (x-x0)dmu).
 *  Weights are normalized as sum(weights)=<1>;
 *  This approach may be less stable than direct calculation from moments.
 *  but it allows us to calculate quadratures for second kind polynomials.
 * @param n The order of generated Radau quadrature is n+1.
 * @param p Orthogonal polynomials, there are should be at least n+2 polynomials.
 */
public double [][] getRadauQuadraturesFromPolynomials(final int n,final double [][] p)
{
    if(n+1>=p.length) throw new RuntimeException("(n="+n+")+1 >= p.length="+p.length);
    // Calculate rr - basis polynomials for Radau measure.
    final double [][] rr=getRadauPolynomials(p);

    final double [] nodes=rr[n];
    final double [][] ab=getAB(rr);
    final double [] xr=getXNodesFromAB(n,ab[0],ab[1]),wr=true?getWeightsChristoffel(xr,rr):getWeights(nodes,xr,rr);
    // convert weigths back to out basis
    final double [] tmp1=new double []{1};tmp1[0]/=fx(tmp1,0);
    final double [] tmp=getRadauMultiplied(tmp1);
    final double x0=-fx(tmp,0)/Dfx(tmp,0);
    final double [] xs=new double [n+1],ws=new double [n+1];
    for(int k=0;k<n;k++){ws[k]=wr[k]/fx(tmp,xr[k]);xs[k]=xr[k];}
    xs[n]=x0;
    final double [] nnodes=new double [nodes.length+1];
    TMatr.expandOverLowerDiagBasis(nnodes,nodes,nodes.length,p);
    final double p00=fx(p[0],0);
    ws[n]=(nnodes[0]/fx(nodes,x0))/p00;

    return new double [][] {xs,ws,getRadauMultiplied(nodes)};
}

/** Generate Radau <(x-x0)Q_k> moments from <Q_k> moments.
 * Rather memory inefficient.
 * Used for unit tests.
 */
public double [] getRadauMomentsFromMoments(final double [] moms)
{
    final double [] r=new double [moms.length-1];
    for(int k=0;k<r.length;k++)
    {
	final double [] tmp=new double [k+1];
	tmp[k]=1;
	r[k]=TMatr.mulVV(k+2,getRadauMultiplied(tmp),moms);
    }
    return r;
}

/** Calcuate Q(x0) moments
 */
public double [] getX0Moments(final int nmoms,final double x0)
{
    final double [] m=new double [nmoms],r=new double [nmoms];
    for(int k=nmoms;--k>=0;)
    {
	r[k]=1;
	m[k]=fx(r,x0);
	r[k]=0;
    }
    return m;
}


/** Generate <xQ_k(x)> moments from <Q_k(x)> moments.
 * Rather memory inefficient.
 * Used for unit tests.
 */
public double [] getXFMomentsFromMoments(final double [] moms)
{
    final double [] r=new double [moms.length-1];
    final double [] rmoms=new double [moms.length];
    for(int k=0;k<r.length;k++)
    {
	final double [] tmp=new double [k+1];
	tmp[k]=1;
	setxP(rmoms,tmp);
	r[k]=TMatr.mulVV(k+2,rmoms,moms);
    }
    return r;
}

/** Generate <P(x) Q_k(x)> moments from <Q_k(x)> moments.
 */
public double [] getPFMomentsFromFMoments(final double [] P,final double [] moms)
{
    final double [] r=new double [moms.length-(P.length-1)],tmp=new double [moms.length];
    for(int j=0;j<r.length;j++)
    {
	double s=0;
	for(int k=0;k<P.length;k++)
	{
	    setQQCoefs(j,k,tmp);
	    s+=P[k]*TMatr.mulVV(1+j+k,moms,tmp);
	}
	r[j]=s;
    }
    return r;
}


/** The returned [][] array contains 3 [] arrays: x nodes, weights and nodes polynomial.
 *  Corresponding to quadratures built on n-th polynomial.
 *  Weights are normalized as sum(weights)=<1>.
 *  Uses generalized eigenvalues expansion. Most stable.
 * @param n Order of the quadrature.
 * @param moments Moments in basis, the array must have at least 2n elements (0..2n-1 moments).
 * @return array, array[0] x-nodes, array[1] weights.
 */
public double [][] getQuadraturesFromMoments_EV(final int n,final double [] moments)
{
    final double [] QQ=getQQMatr(n,moments),QQx=getQQMatr(n,getXFMomentsFromMoments(moments));
    //System.err.println("QQ="+java.util.Arrays.toString(QQ)+"\nQQx="+java.util.Arrays.toString(QQx)+"\nmoments="+java.util.Arrays.toString(moments));
    final double [] ev=EV.getGEV_Generalized(n,QQx,n,QQ,n,true);
    final double [] x=new double [n],w=new double [n],psi=new double [n],nodes=new double [n+1];
    nodes[0]=1;
    for(int k=0;k<n;k++)
    {
	x[k]=ev[k];
	System.arraycopy(ev,n+k*n,psi,0,n);
	w[k]=1/Math.pow(fx(psi,x[k]),2);
	if(k==0) {setxP(nodes,psi);TMatr.sumP1(nodes,psi,psi.length,-x[k]);}
    }
    return new double [][]{x,w,nodes};
}

/** Similar to getQuadraturesFromMoments_EV
 * but calculates quadratures directly: find n-th orthogonal polynomial,
 * find roots, then weights.
 * Used for accuracy estimation of getQuadraturesFromMoments_EV.
 * @param n Order of the quadrature.
 * @param moments Moments in basis, must be at least 2n elements (0..2n-1 moments).
 * @return array, array[0] x-nodes, array[1] weights, array[2] nodes polynomial.
 */
public double [][] getQuadraturesFromMoments_Direct(final int n,final double [] moments)
{
    final double [] P=getSingleOrthogonalPolynomial(n,moments);
    final double [] x=getXNodes(P);
    final double [] w=getWeightsByIntegration(P,x,moments);
    return new double [][]{x,w,P};
}



/** The returned [][] array contains 2 [] arrays: x nodes and weights.
 *  Corresponding to quadratures built on n-th polynomial.
 *  Weights are normalized as sum(weights)=<1>.
 *  Uses eigenvectors expansion in orthogonal polynomials basis.
 *  Most stable.
 * @param n Quadrature order.
 * @param a Recurrence coefficient, must have at least n+1 element.
 * @param b Recurrence coefficient must have at least n elements.
 * @return The quadrature.
 */
public double [][] getQuadraturesFromAB_EV(final int n,final double [] a,final double [] b)
{
    final double [] ev=EV.getEV_Thridiagonal(n,b,0,a,1,true);
    final double [] x=new double [n],w=new double [n],psi=new double [n];
    for(int k=0;k<n;k++)
    {
	x[k]=ev[k];
	System.arraycopy(ev,n+k*n,psi,0,n);
	w[k]=1/Math.pow(getSUMQxFromAB(psi,x[k],a,b),2);
    }
    return new double [][]{x,w};
}


/** The returned [][] array contains 3 [] arrays: x nodes, weights and nodes polynomial.
 *  Corresponding to quadratures built on n+1 -th Radau polynomial.
 *  Weights are normalized as sum(weights)=<1>.
 *  Uses generalized eigenvalues expansion. Most stable.
 * @param n The order of generated Radau quadrature is n+1.
 * @param moments Moments, should be at least 2n+1 elements.
 * @return quadrature.
 */
public double [][] getRadauQuadraturesFromMoments_EV(final int n,final double [] moments)
{
    final double [][] nn=getQuadraturesFromMoments_EV(n,getRadauMomentsFromMoments(moments));
    final double [] xr=nn[0],wr=nn[1],nodes=nn[2];
    final double [] tmp1=new double []{1};tmp1[0]/=fx(tmp1,0);
    final double [] tmp=getRadauMultiplied(tmp1);
    final double x0=-fx(tmp,0)/Dfx(tmp,0);
    final double [] xs=new double [n+1],ws=new double [n+1];
    for(int k=0;k<n;k++){ws[k]=wr[k]/fx(tmp,xr[k]);xs[k]=xr[k];}
    xs[n]=x0;
    ws[n]=TMatr.mulVV(nodes.length,nodes,moments)/fx(nodes,x0);

    nn[0]=xs;nn[1]=ws;nn[2]=getRadauMultiplied(nodes);
    return nn;
}



/** Given <Q_i|Q_j> matrix calculate first n orthonormal vectors
 *  using Gramm Schmidt orthogonalization.
 *  In polynomial basis gives first n orthogonal polynomials.
 *  Note: may be numerically unstable.
 *  @param n The number of basis vectors to generate.
 *  @param QQ The <Q_i|Q_j> matrix, generated by getQQMatr() method from moments, i,j=0..n-1.
 *  @param nm The leading dimension of QQ.
 *  @return Othogonal basis obtained by applying Gramm Schmidt orthogonalization
 *  to basis vectors in original order.
 */
public static double [][] getGrammSchmidtOrthogonalized(final int n,final double [] QQ,final int nm)
{
    if(!(n<=nm)) throw new IllegalArgumentException("n="+n+" nm="+nm);
    final double [][] basis=new double [n][];

    for(int i=0;i<n;i++)
    {
	final double [] b=basis[i]=new double [i+1];
	b[i]=1;
	for(int k=i,inm=i*nm;--k>=0;)
	{
	    TMatr.sumP1(b,basis[k],k+1,-TMatr.mulVV(k+1,QQ,inm,basis[k],0));
	}
	// normalize
	final double anorm=Math.sqrt(TMatr.getAMatrA(i+1,b,QQ,nm));
	for(int k=i;k>=0;--k){b[k]/=anorm;}
    }
    return basis;
}


/** Calculates single Type II multiple orthogonal polynomial.
 *  @param multiindex Multiindex. Typically all elements are the same.
 *  @param measures Moment measures <Q_k>_mu_s. Each s-th measure should have at least sum(multiindex)+multiindex[s] elements.
 *  @return Type II multiple orthogonal polynomial.
 */
public double [] getMultipleOrthogonalPolynomial_TypeII(final int [] multiindex,final double [][] measures)
{
    int N=0;
    for(int k=multiindex.length;--k>=0;) {N+=multiindex[k];}
    final double [] P=new double [N+1],matr=new double [N*N],tmp=new double [N*2+1];
    for(int j=0;j<=N;j++)
    {
	for(int i=0,m=0;m<multiindex.length;m++)
	{
	    final int im=multiindex[m];
	    for(int k=0;k<im;k++,i++)
	    {
		setQQCoefs(j,k,tmp);
		final double mom=TMatr.mulVV(1+j+k,measures[m],tmp);
		if(j<N)
		{
		    matr[i*N+j]=mom;
		}
		else
		{
		    P[i]=-mom;
		}
	    }
	}
    }
    final int [] ipvt=new int [N];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,N,N,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(P,Double.NaN);
	return P;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,N,N,ipvt,P,1);
    P[N]=1;
    return P;
}

/** Calculates Gaussian quadratures for multiple orthogonal polynimials.
 *  @param multiindex Multiindex, typically all elements are the same.
 *  @param measures Moment measures <Q_k>_mu_s. Each s-th measure should have at least sum(multiindex)+multiindex[s] elements.
 *  @return Quadratures. The array has 3+multiindex.length elements.
 *  0-th element: x nodes.
 *  1-st element: imag(x), all equal to 0 on success, complex roots indicate failure to find proper quadrature.
 *  s+2  element: weight[s], s=0..multiindex.length-1;
 *  multiindex.length-1+3-th element: the nodes polynomial in basis.
 *  The results tested to match exactly the
 *  Carlos F. Borges,Numer. Math. 67: 271-288 (1994),
 *  On a class of Gauss-like quadrature rules.
 *  However the results are not stable for arbitrary measures,
 *  the method TestMonomials.test_MultipleOrthogonality() fails for some randomly generated measures.
 */
public double [][] getQuadraturesForMultipleOrthogonalPolynomial(final int [] multiindex,final double [][] measures)
{
    final double [] P=getMultipleOrthogonalPolynomial_TypeII(multiindex,measures);
    final int N=P.length-1;
    final double [][] w=new double [3+multiindex.length][];
    final double [] rex=w[0]=new double [N],imx=w[1]=new double [N],p=new double [P.length-1];
    w[w.length-1]=P;
    final int nr=getPolynomialRootsFinderInBasis().findRoots(rex.length,P,rex,imx);
    if(nr!=N)
    {
	java.util.Arrays.fill(rex,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(rex.length,rex,imx);
    }
    final int s_offset=2;
    for(int m=0;m<multiindex.length;m++)
    {
	w[m+s_offset]=getWeightsByIntegration(P,rex,measures[m]);
    }
    return w;
}



/** Calculates Gauss-Kronrod quadratures.
 *  @param n Order or original Gauss Quadrature, final quadrature has 2n+1 order.
 *  @param moms Moments, must have at least 3n+2 elements (0..3n+1 moments).
 *  @return Quadratures. The array has 6 elements.
 *  0-th element: Gauss x nodes.
 *  1-st element: imag(x) of Gauss nodes, all equal to 0 on success.
 *  2-nd element: weights for Gauss nodes.
 *  3-rd element: Kronrod x nodes.
 *  4-th element: imag(x) of Kronrod nodes, all equal to 0 on success.
 *  5-th element: weights for Kronrod nodes.
 */
public double [][] getKronrodQuadratures(final int n,final double [] moms)
{
    final PolynomialRootsFinder rf=getPolynomialRootsFinderInBasis();

    // Gauss nodes
    final double [] P=getSingleOrthogonalPolynomial(n,moms);
    final double [] gxre=new double [n],gxim=new double [n];
    final int ng=rf.findRoots(gxre.length,P,gxre,gxim);
    if(ng!=n)
    {
	java.util.Arrays.fill(gxre,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(n,gxre,gxim);
    }

    // Kronrod nodes
    final double [] momsP=getPFMomentsFromFMoments(P,moms);
    final double [] K=getSingleOrthogonalPolynomial(n+1,momsP);
    final double [] kxre=new double [n+1],kxim=new double [n+1];
    final int nk=rf.findRoots(kxre.length,K,kxre,kxim);
    if(nk!=(n+1))
    {
	java.util.Arrays.fill(kxre,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(n+1,kxre,kxim);
    }

    // Weights for both: Gauss & Kronrod nodes.
    final double [] momsK=getPFMomentsFromFMoments(K,moms);
    final double [] wg=getWeightsByIntegration(P,gxre,momsK),
	wk=getWeightsByIntegration(K,kxre,momsP);
    for(int k=0;k<gxre.length;k++){wg[k]/=fx(K,gxre[k]);}
    for(int k=0;k<kxre.length;k++){wk[k]/=fx(P,kxre[k]);}

    return new double [][]{gxre,gxim,wg,kxre,kxim,wk};
}

/** Multiplies P*Q.
 */
public double [] getPQMultiplied(final double [] P,final double [] Q)
{
    final double [] res=new double [P.length+Q.length-1];
    B.setMult2Pol(res,P,P.length-1,Q,Q.length-1,new double [res.length]);
    return res;
}

/** Expand Q_{l}Q_{m}=sum_{j=0}^{j=l+m}Q_{j}coefs[j].
 *  Saves 1+l+m coefficients to coefs, index range 0..l+m.
 */
public void setQQCoefs(final int l,final int m,final double [] coefs)
{
    B.setQlQmExpansionCoefs(l,m,coefs);
}

/** Multiply P(x)=\sum_{k=0}^{k=n} p[k] Q_k(x) by x in Q_k(x) basis.
 *  The result is saved to res, which must have at least p.length+1 elements available.
 */
public void setxP(final double [] res,final double [] p)
{
    B.setaxbP(p.length-1,res,p,1,0);
}

/** Synthetically divide P(x)=\sum_{k=0}^{k=n} p[k] Q_k(x) by x-x0 in Q_k(x) basis.
 *  The remainder is equal to returned value.
 *  The quotient is saved to res, which must have at least p.length-1 elements available.
 */
public double sdiv1(final double [] res,final double [] p,final double x0)
{
    return B.sdiv1(p.length-1,res,p,x0);
}

/** Calculate f(x)=\sum_{k=0}^{k=n} m[k] Q_k(x).
 */
public double fx(final double [] m,final double x)
{
    return B.calculateSumD0(m.length-1,m,x);
}

/** Calculate f'(x)=\sum_{k=0}^{k=n} m[k] Q'_k(x).
 */
public double Dfx(final double [] m,final double x)
{
    return B.calculateSumD1(m.length-1,m,x);
}

/** Calculate f''(x)=\sum_{k=0}^{k=n} m[k] Q''_k(x).
 */
public double D2fx(final double [] m,final double x)
{
    return B.calculateSumD2(m.length-1,m,x);
}

/** Polynomial root finder in original basis.
 *  Used for unit testing the quadratures.
 */
public PolynomialRootsFinder getPolynomialRootsFinderInBasis()
{
    return new PolynomialRootsConfederateMatrixABasis.PolynomialRootsConfederateMatrixPBasis(B);
}


/** Multiply nodes0 polynomial, P(x)=\sum_{k=0}^{k=n} nodes0[k] Q_k(x),
 *  by (x-x0) or (x0-x) , whatever is positive
 *  to obtain Radau nodes polynomial.
 *  The result is saved to res, which must have at least nodes0.length+1 elements available.
 *  All calculations are performed in the basis Q_k(x) used in this class.
 *  Can be set as a combination of setxP and nodes0,
 *  or calculated directly.
 */
public double [] getRadauMultiplied(final double [] nodes0)
{
    throw new RuntimeException("The getRadauMultiplied() must be implemented if getRadauQuadraturesFromPolynomials() is used");
}

/** Calculate the root of getRadauMultiplied([1]).
 */
public double getRadauRoot()
{
    final double [] rtmp=getRadauMultiplied(new double []{1});
    return -fx(rtmp,0)/Dfx(rtmp,0);
}



/** Used for diagnostics. */
public static void printQuadratures(final String txt,final int n,final double [] moments,final OrthogonalPolynomialsABasis q,final boolean flag_printab)
{
    final double [][] qg=q.getQuadraturesFromMoments_EV(n,moments),
	qr=q.getRadauQuadraturesFromMoments_EV(n-1,moments); // Note: Radu nodes have one extra.
    final double [] xg=qg[0],wg=qg[1],xr=qr[0],wr=qr[1];
    SortArrays.sortArrayWithLoad(xg.length,xg,wg);
    SortArrays.sortArrayWithLoad(xr.length,xr,wr);
    if(xr.length!=xg.length) throw new RuntimeException("Mismatch");

    System.err.println("========="+txt+": Gauss quadratures for n_nodes="+n);
    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<n;k++)
    {
	s+=wg[k];
	System.err.println(String.valueOf(xg[k])+"\t"+wg[k]+"\t"+s);
    }
    //UnitTests.checkDoubleRel(s,"s",1,"1",eps);

    System.err.println("========="+txt+": Radau quadratures for n_nodes="+n);
    System.err.println("X\tW\tS");
    s=0;
    for(int k=0;k<n;k++)
    {
	s+=wr[k];
	System.err.println(String.valueOf(xr[k])+"\t"+wr[k]+"\t"+s);
    }
    //UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    if(flag_printab)
    {
	System.err.println();
	final double [][] ab=q.getAB(q.getOrthogonalPolynomialsFirstKind(n,moments));
	final double [] a=ab[0],b=ab[1];
	System.err.println("K\tA\tB");
	for(int k=0;k<n;k++)
	{
	    System.err.println(String.valueOf(k)+"\t"+(k<a.length?a[k]:"N/A")+"\t"+(k<b.length?b[k]:"N/A"));
	}
    }

    System.err.println();
    System.err.println();
}



}
