package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Fortran polynomial roots finder
 *  http://www.netlib.org/toms/493
 *  translated to java with few fixes:
 *  1. scaling is broken in original version, new function pscale,
 *  adopted from http://www.netlib.org/toms/419 is used.
 *  2. some conditions are replaced to avoid infinite loops with NaNs.
 */
public class PolynomialRootJenkinsTraubORIG  implements PolynomialRootsFinder {


 private static final  double
     DBL_EPSILON=Math.ulp(1.0),
     DBL_MIN=Double.MIN_NORMAL,
     DBL_MAX=Double.MAX_VALUE,
     COSR = Math.cos((94.0/180.0)*Math.PI),
     SINR = Math.sin((94.0/180.0)*Math.PI),
     SQRT2D2=0.5*Math.sqrt(2.0)
    ;
    private static final double
/*
 THE FOLLOWING STATEMENTS SET MACHINE CONSTANTS USED
 IN VARIOUS PARTS OF THE PROGRAM. THE MEANING OF THE
 FOUR CONSTANTS ARE...
 ETA     THE MAXIMUM RELATIVE REPRESENTATION ERROR
         WHICH CAN BE DESCRIBED AS THE SMALLEST
         POSITIVE FLOATING POINT NUMBER SUCH THAT
         1.D0+ETA IS GREATER THAN 1.
 INFIN   THE LARGEST FLOATING-POINT NUMBER.
 SMALNO  THE SMALLEST POSITIVE FLOATING-POINT NUMBER
         IF THE EXPONENT RANGE DIFFERS IN SINGLE AND
         DOUBLE PRECISION THEN SMALNO AND INFIN
         SHOULD INDICATE THE SMALLER RANGE.
 BASE    THE BASE OF THE FLOATING-POINT NUMBER
         SYSTEM USED.
 THE VALUES BELOW CORRESPOND TO THE BURROUGHS B6700
*/
	BASE=2,
	LOG_BASE=Math.log(BASE),
	ETA=DBL_EPSILON,
	INFIN=DBL_MAX,
	SMALNO=DBL_MIN,
/*
 ARE AND MRE REFER TO THE UNIT ERROR IN + AND *
 RESPECTIVELY. THEY ARE ASSUMED TO BE THE SAME AS
 ETA.
*/
	ARE=ETA,
	MRE=ETA,
	HI=Math.sqrt(INFIN),
	LO=SMALNO/ETA;


    private static final int _TRACE=0;

    private final int MAXDEGREE;
    private final double [] P, QP, K, QK, SVK, PT, TEMP;
    public PolynomialRootJenkinsTraubORIG(final int MAXDEGREE)
    {
	this.MAXDEGREE=MAXDEGREE;
	final int s=MAXDEGREE+2;
	this.P=new double [s];
	this.QP=new double [s];
	this.K=new double [s];
	this.QK=new double [s];
	this.SVK=new double [s];
	this.PT=new double [s];
	this.TEMP=new double [s];
    }


    private double
	SR, SI, U, V, A, B, C, D,
	A1, A2, A3, A7, E, F, G, H,
	SZR, SZI, LZR, LZI;

/**
 FINDS THE ZEROS OF A REAL POLYNOMIAL
 OP  - DOUBLE PRECISION VECTOR OF COEFFICIENTS IN
       ORDER OF INCREASING POWERS.
 DEGREE   - INTEGER DEGREE OF POLYNOMIAL.
 ZEROR, ZEROI - OUTPUT DOUBLE PRECISION VECTORS OF
                REAL AND IMAGINARY PARTS OF THE
                ZEROS.
*/
public int findRoots(final int degree,
		     final double [] op,
		     final double [] zeror,
		     final double [] zeroi)
{

if(_TRACE>0){
    System.err.println("RPOLY: START N="+degree);
}

if (degree > MAXDEGREE)
{
    // The entered Degree is greater than MAXDEGREE. Exiting rpoly. No further action taken.
    throw new RuntimeException("degree="+degree+" > MAXDEGREE="+MAXDEGREE);
}


// ALGORITHM FAILS IF THE LEADING COEFFICIENT IS ZERO.
if (!(op[degree-0] != 0))
{
    return -1;
}

int N=degree;
// REMOVE THE ZEROS AT THE ORIGIN IF ANY
for(int j=0; N>=0 && op[degree-N] == 0 ; j++)
{
    zeror[j] = zeroi[j] = 0.0;
    N--;
}

int NN = N + 1;

// MAKE A COPY OF THE COEFFICIENTS.
// Reverse the order so polinomial powers go in INCREASING orders
// in contrast with original RPOLY, also shift by 1 for matching
// fortran style of arrays
for (int i = 0; i < NN; i++)   P[i+1] = op[degree-i];


double xx = SQRT2D2;
double yy = -xx;

mloop:while (N >= 1){ // Main loop
    // START THE ALGORITHM FOR ONE ZERO
    if (N <= 2){
	// CALCULATE THE FINAL ZERO OR PAIR OF ZEROS
	if (N < 2){
	    zeror[degree-1] = -P[2]/P[1];
	    zeroi[degree-1] = 0.0;
	} else {
	    quad(P[1], P[2], P[3]);
	    zeror[degree-2]=this.SZR;
	    zeroi[degree-2]=this.SZI;
	    zeror[degree-1]=this.LZR;
	    zeroi[degree-1]=this.LZI;
	}
	//return degree;
	N=0;
	break mloop;
    }


    // scale if necessary
    if(true) {
	final double factor = pscale(NN,P);
	if(factor!=1)
	{
	    for(int i=1;i<=NN;i++) P[i]*=factor;
	}
    } else {
	final double f = Math.abs(P[1]);
	if(!(f==0))
	{
	    final double factor=Math.pow(BASE,(int)(Math.log(f)/LOG_BASE + 0.5));
	    for(int i=1;i<=NN;i++) P[i]/=factor;
	}
    }

    // COMPUTE LOWER BOUND ON MODULI OF ZEROS.
    for(int i=1;i<=NN;i++) PT[i]=Math.abs(P[i]);
    PT[NN]=-PT[NN];

    // COMPUTE UPPER ESTIMATE OF BOUND
    double x = Math.exp((Math.log(-PT[NN])-Math.log(PT[1]))/N);
    if(PT[N]!=0)
    {
	// IF NEWTON STEP AT THE ORIGIN IS BETTER, USE IT.
	final double xm = -PT[NN]/PT[N];
	if (xm<x) x = xm;
	if(_TRACE>0){
	    System.err.println("RPOLY: NEWTON STEP X="+x);
	}
    }

    // CHOP THE INTERVAL (0,X) UNTIL FF <= 0
    int ni=0;
    for(double xm;ni<1000;ni++) {
	xm=x*0.1;
	double ff = PT[1];
	for(int i=2;i<=NN;i++)
	{
	    ff=ff*xm+PT[i];
	}
	if(!(ff>0) || !(xm>SMALNO)) break;
	x = xm;
    }
    if(_TRACE>0){
	System.err.println("RPOLY: CHOP X="+x);
    }


    // DO NEWTON ITERATION UNTIL X CONVERGES TO TWO DECIMAL PLACES
    ni=0;
    for(double dx=x;Math.abs(dx/x)>0.005 && ni<33;ni++)
    {
	if(_TRACE>0){
	    System.err.println("RPOLY: NEWTON ITERATION X="+x+" DX="+dx);
	}
        double ff = PT[1],df = ff;
	for(int i=2;i<=N;i++)
	{
	    ff = ff*x + PT[i];
	    df = df*x + ff;
	}
	ff = ff*x + PT[NN];
	dx = ff/df;
	x = x - dx;
    }
    final double bnd = x;
    if(_TRACE>0){
	System.err.println("RPOLY: BND="+bnd);
    }

    // COMPUTE THE DERIVATIVE AS THE INTIAL K POLYNOMIAL AND DO 5 STEPS WITH NO SHIFT
    final int nm1 = N - 1;
    for(int i=2;i<=N;i++)
    {
        K[i]=(NN-i)*P[i]/N;
    }
    K[1]=P[1];
    final double aa = P[NN],bb = P[N];
    boolean zerok = K[N]==0;
    for(int jj=1;jj<=5;jj++)
    {
        final double cc = K[N];
	if(_TRACE>0){
	    System.err.println("RPOLY: 5 Step JJ="+jj+" CC="+cc+
			       " AA="+aa+" BB="+bb+" ZEROK="+zerok);
	}
        if(!zerok)
	{
	    // USE SCALED FORM OF RECURRENCE IF VALUE OF K AT 0 IS NONZERO
	    final double t = -aa/cc;
	    for(int i=1;i<=nm1;i++)
	    {
		final int j = NN - i;
		K[j] = t*K[j-1]+P[j];
	    }
	    K[1]=P[1];
	    zerok = Math.abs(K[N])<=Math.abs(bb)*ETA*10.0;
	}
	else
	{
	    // USE UNSCALED FORM OF RECURRENCE
	    for(int i=1;i<=nm1;i++)
	    {
		final int j = NN - i;
		K[j] = K[j-1];
	    }
	    K[1]=0;
	    zerok = K[N]==0;
	}
    }
    // SAVE K FOR RESTARTS WITH NEW SHIFTS
    for(int i=1;i<=N;i++)
    {
	TEMP[i]=K[i];
    }
    if(_TRACE>0){
	System.err.println("RPOLY: NEW SHIFTS");
    }
    // LOOP TO SELECT THE QUADRATIC  CORRESPONDING TO EACH NEW SHIFT
    for(int cnt=1;cnt<=20;cnt++)
    {
	// QUADRATIC CORRESPONDS TO A DOUBLE SHIFT TO A
	// NON-REAL POINT AND ITS COMPLEX CONJUGATE. THE POINT
	// HAS MODULUS BND AND AMPLITUDE ROTATED BY 94 DEGREES
	// FROM THE PREVIOUS SHIFT
        final double xxx = COSR*xx - SINR*yy;
	yy = SINR*xx + COSR*yy;
        xx = xxx;
        this.SR = bnd*xx;
        this.SI = bnd*yy;
        this.U = -2*this.SR;
        this.V = bnd;
	// SECOND STAGE CALCULATION, FIXED QUADRATIC
        final int nz=fxshfr(N,NN,20*cnt);
	if(_TRACE>0){
	    System.err.println("RPOLY: CNT="+cnt+" NZ="+nz);
	    dumpc("AFTER FXSHFR");
	}
        if (nz>0)
	{
	    // THE SECOND STAGE JUMPS DIRECTLY TO ONE OF THE THIRD
	    // STAGE ITERATIONS AND RETURNS HERE IF SUCCESSFUL.
	    // DEFLATE THE POLYNOMIAL, STORE THE ZERO OR ZEROS AND
	    // RETURN TO THE MAIN ALGORITHM.
	    final int j = degree - N + 1;
	    zeror[j-1] = this.SZR;
	    zeroi[j-1] = this.SZI;
	    NN = NN - nz;
	    N = NN - 1;
	    for(int i=1;i<=NN;i++)
	    {
		P[i]=QP[i];
	    }
	    if(nz>1)
	    {
		zeror[j]=this.LZR;
		zeroi[j]=this.LZI;
	    }
	    continue mloop;
	}
	// IF THE ITERATION IS UNSUCCESSFUL ANOTHER QUADRATIC
	// IS CHOSEN AFTER RESTORING K
	for(int i=1;i<=N;i++)
	{
	    K[i]=TEMP[i];
	}
    }
    // RETURN WITH FAILURE IF NO CONVERGENCE WITH 20 SHIFTS
    break mloop;
} // end mloop

return degree-N;

}


/**
 COMPUTES UP TO  L2  FIXED SHIFT K-POLYNOMIALS,
 TESTING FOR CONVERGENCE IN THE LINEAR OR QUADRATIC
 CASE. INITIATES ONE OF THE VARIABLE SHIFT
 ITERATIONS AND RETURNS WITH THE NUMBER OF ZEROS
 FOUND.
 L2 - LIMIT OF FIXED SHIFT STEPS
 @returns nz - number of zeros found
*/
private int fxshfr(final int N,final int NN,final int l2)
{
      double betav = .25;
      double betas = .25;
      double oss = this.SR;
      double ovv = this.V;
      double otv=Double.NaN, ots=Double.NaN;
      // EVALUATE POLYNOMIAL BY SYNTHETIC DIVISION
      quadsd(NN, this.U, this.V, P, QP);
      this.A=quadsd_A;
      this.B=quadsd_B;
      int type=calcsc(N);
      if(_TRACE>0){
	      System.err.println("FXSHFR: INIT TYPE="+type+
				 " A="+this.A+" B="+this.B+
				 " K[N]="+K[N]+
				 " K[N-1]="+K[N-1]);
	      dumpc("FXSHFR: INIT");
      }
      for(int j=1;j<=l2;j++)
      {
	  // CALCULATE NEXT K POLYNOMIAL AND ESTIMATE V
	  nextk(N,type);
	  type=calcsc(N);
	  newest(N,NN,type); // sets newest_uu & newest_vv
	  double ui=newest_uu;
	  double vi=newest_vv;
	  if(_TRACE>0){
	      System.err.println("FXSHFR: J="+j+" TYPE="+type+
				 " UI="+ui+" VI="+vi+
				 " K[N]="+K[N]+
				 " K[N-1]="+K[N-1]);
	      dumpc("L2 loop");
	  }
	  final double vv = vi;
	  // ESTIMATE S
	  final double ss = K[N]!=0?-P[NN]/K[N]:0;
	  double tv = 1;
	  double ts = 1;
      l70:if(j>1 && type!=3)
	  {
	      // COMPUTE RELATIVE MEASURES OF CONVERGENCE OF S AND V SEQUENCES
	      if (vv!=0) tv = Math.abs((vv-ovv)/vv);
	      if (ss!=0) ts = Math.abs((ss-oss)/ss);
	      // IF DECREASING, MULTIPLY TWO MOST RECENT CONVERGENCE MEASURES
	      final double tvv = tv<otv?tv*otv:1;
	      final double tss = ts<ots?ts*ots:1;
	      // COMPARE WITH CONVERGENCE CRITERIA
	      final boolean vpass = tvv<betav;
	      final boolean spass = tss<betas;
	      if (! (spass || vpass)) break l70;

	      // AT LEAST ONE SEQUENCE HAS PASSED THE CONVERGENCE TEST.
	      // STORE VARIABLES BEFORE ITERATING
	      final double svu = this.U;
	      final double svv = this.V;
	      for(int i=1;i<=N;i++)
	      {
		  SVK[i]=K[i];
	      }
	      double s = ss;
	      // CHOOSE ITERATION ACCORDING TO THE FASTEST CONVERGING SEQUENCE
	      boolean vtry = false;
	      boolean stry = false;
	  l20:for(boolean l40=!(spass && (!vpass || tss<tvv));;)
	      {
	      l50:{
		      if(l40)
		      {
			  final int nz=quadit(N,NN,ui,vi);
			  if(_TRACE>0){
			      System.err.println("FXSHFR: QUADIT NZ="+nz);
			  }
			  if (nz>0) return nz;
			  // QUADRATIC ITERATION HAS FAILED. FLAG THAT IT HAS
			  // BEEN TRIED AND DECREASE THE CONVERGENCE CRITERION.
			  vtry = true;
			  betav = betav*0.25;
			  // TRY LINEAR ITERATION IF IT HAS NOT BEEN TRIED AND
			  // THE S SEQUENCE IS CONVERGING
			  if(stry || !spass) { break l50;}
			  for(int i=1;i<=N;i++)
			  {
			      K[i]=SVK[i];
			  }
		      }
		      l40=true;

		      final int nz=realit(N,NN,s);
		      if(_TRACE>0){
			  System.err.println("FXSHFR: REALIT NZ="+nz);
		      }
		      if (nz>0) return nz;
		      // LINEAR ITERATION HAS FAILED. FLAG THAT IT HAS BEEN
		      // TRIED AND DECREASE THE CONVERGENCE CRITERION
		      stry = true;
		      betas = betas*0.25;
		      s=realit_sss;
		      if(realit_iflag==0){ break l50;}
		      // IF LINEAR ITERATION SIGNALS AN ALMOST DOUBLE REAL
		      // ZERO ATTEMPT QUADRATIC INTERATION
		      ui = -(s+s);
		      vi = s*s;
		      continue l20;
		  } // end l50
		  // RESTORE VARIABLES
		  this.U = svu;
		  this.V = svv;
		  for(int i=1;i<=N;i++)
		  {
		      K[i]=SVK[i];
		  }
		  // TRY QUADRATIC ITERATION IF IT HAS NOT BEEN TRIED
		  // AND THE V SEQUENCE IS CONVERGING
		  if (!vpass || vtry) break l20;
	      } // end l20
	      // RECOMPUTE QP AND SCALAR VALUES TO CONTINUE THE SECOND STAGE
	      quadsd(NN, this.U, this.V, P, QP);
	      this.A=quadsd_A;
	      this.B=quadsd_B;
	      type=calcsc(N);
	  } // end l70
	  ovv = vv;
	  oss = ss;
	  otv = tv;
	  ots = ts;
      } // end j=1..l2 loop
      return 0; // not found
}


/**
   VARIABLE-SHIFT K-POLYNOMIAL ITERATION FOR A
   QUADRATIC FACTOR CONVERGES ONLY IF THE ZEROS ARE
   EQUIMODULAR OR NEARLY SO.
   UU,VV - COEFFICIENTS OF STARTING QUADRATIC
   NZ - NUMBER OF ZERO FOUND
*/
private int quadit(final int N,final int NN,final double uu,final double vv)
{
    boolean tried = false;
    this.U = uu;
    this.V = vv;
    double relstp=Double.NaN,omp=Double.NaN;
    for(int j=0;;)
    {
	// MAIN LOOP
	quad(1.0, this.U, this.V); //SZR, SZI, LZR, LZI are set
	// RETURN IF ROOTS OF THE QUADRATIC ARE REAL AND NOT
	// CLOSE TO MULTIPLE OR NEARLY EQUAL AND  OF OPPOSITE SIGN
	if(!(Math.abs(Math.abs(this.SZR)-Math.abs(this.LZR))<=0.01*Math.abs(this.LZR)))
	{
	    break;
	}
	// EVALUATE POLYNOMIAL BY QUADRATIC SYNTHETIC DIVISION
	quadsd(NN,this.U,this.V,P,QP);
	this.A=quadsd_A;
	this.B=quadsd_B;
	final double mp = Math.abs(this.A-this.SZR*this.B) + Math.abs(this.SZI*B);

	// COMPUTE A RIGOROUS  BOUND ON THE ROUNDING ERROR IN EVALUTING P
	final double zm = Math.sqrt(Math.abs(this.V));
	double ee = 2.0*Math.abs(QP[1]);
	final double t = -this.SZR*this.B;
	for(int i=2;i<=N;i++)
	{
	    ee=ee*zm+Math.abs(QP[i]);
	}
	ee = ee*zm + Math.abs(this.A+t);
	ee = (5.0*MRE+4.0*ARE)*ee - (5.0*MRE+2.0*ARE)*
	    (Math.abs(this.A+t)+Math.abs(this.B)*zm) + 2.0*ARE*Math.abs(t);
	if(_TRACE>0)
	{
	    System.err.println("quadit: N="+N+" ee="+ee+" t="+t);
	}
	// ITERATION HAS CONVERGED SUFFICIENTLY IF THE
	// POLYNOMIAL VALUE IS LESS THAN 20 TIMES THIS BOUND
	if(mp<=20*ee)
	{
	    if(_TRACE>0)
	    {
		System.err.println("quadit: GOOD EXIT: j="+j+" mp="+mp+" ee="+ee+
				   " zm="+zm+" t="+t+
				   " A="+this.A+" B="+this.B+
				   " SZR="+this.SZR+" SZI="+this.SZI+
				   " LZR="+this.LZR+" LZI="+this.LZI);
	    }
	    return 2;
	}
	// STOP ITERATION AFTER 20 STEPS
	if(++j>20) break;
	if(j>=2 && relstp<=0.01 && mp>=omp && !tried)
	{
	    // A CLUSTER APPEARS TO BE STALLING THE CONVERGENCE.
	    // FIVE FIXED SHIFT STEPS ARE TAKEN WITH A U,V CLOSE
	    // TO THE CLUSTER
	    if (relstp<ETA) relstp = ETA;
	    relstp = Math.sqrt(relstp);
	    this.U = this.U - this.U*relstp;
	    this.V = this.V + this.V*relstp;
	    quadsd(NN, this.U, this.V, P, QP);
	    this.A=quadsd_A;
	    this.B=quadsd_B;
	    for(int i=1;i<=5;i++)
	    {
		final int type=calcsc(N);
		nextk(N,type);
	    }
	    tried = true;
	    j = 0;
	}
	omp = mp;
	// CALCULATE NEXT K POLYNOMIAL AND NEW U AND V
	int type=calcsc(N);
	nextk(N,type);
	type=calcsc(N);
	newest(N,NN,type); // sets newest_uu & newest_vv
	final double ui=newest_uu,vi=newest_vv;
	// IF VI IS ZERO THE ITERATION IS NOT CONVERGING
	if (!(vi!=0)) break;
	relstp = Math.abs((vi-this.V)/vi);
	this.U = ui;
	this.V = vi;
    }
    return 0;
}

private int realit_iflag=0;
private double realit_sss=0;
/** VARIABLE-SHIFT H POLYNOMIAL ITERATION FOR A REAL ZERO.
    SSS   - STARTING ITERATE
    realit_iflag - FLAG TO INDICATE A PAIR OF ZEROS NEAR REAL AXIS.
    @return NUMBER OF ZERO FOUND
*/
private int realit(final int N,final int NN,final double sss)
{
    realit_iflag=0;

    final int nm1 = N - 1;
    double s=realit_sss=sss,omp=Double.NaN,t=Double.NaN;
    // MAIN LOOP
    for(int j=0;;)
    {
	double pv = P[1];
	// EVALUATE P AT S
	QP[1] = pv;
	for(int i=2;i<=NN;i++)
	{
	    pv = pv*s + P[i];
	    QP[i] = pv;
	}
	final double mp = Math.abs(pv);
	// COMPUTE A RIGOROUS BOUND ON THE ERROR IN EVALUATING P
	final double ms = Math.abs(s);
	double ee = (MRE/(ARE+MRE))*Math.abs(QP[1]);
	for(int i=2;i<=NN;i++)
	{
	    ee = ee*ms + Math.abs(QP[i]);
	}
	// ITERATION HAS CONVERGED SUFFICIENTLY IF THE
	// POLYNOMIAL VALUE IS LESS THAN 20 TIMES THIS BOUND
	if(mp<=20.0*((ARE+MRE)*ee-MRE*mp))
	{
	    this.SZR = s;
	    this.SZI = 0.0;
	    if(_TRACE>0)
	    {
		System.err.println("realit: GOOD EXIT: j="+j+" mp="+mp+" ee="+ee+
				   " t="+t+" A="+this.A+" B="+this.B+
				   " SZR="+this.SZR+" SZI="+this.SZI);
	    }
	    return 1;
	}
	// STOP ITERATION AFTER 10 STEPS
	if (++j>10) break;
	if (j>=2 && Math.abs(t)<=0.001*Math.abs(s-t) && mp>omp)
	{
	    // A CLUSTER OF ZEROS NEAR THE REAL AXIS HAS BEEN
	    // ENCOUNTERED RETURN WITH IFLAG SET TO INITIATE A
	    // QUADRATIC ITERATION
	    realit_iflag = 1;
	    realit_sss = s;
	    return 0;
	}
	// RETURN IF THE POLYNOMIAL VALUE HAS INCREASED SIGNIFICANTLY
	omp = mp;
	// COMPUTE T, THE NEXT POLYNOMIAL, AND THE NEW ITERATE
	double kv = K[1];
	QK[1] = kv;
	for(int i=2;i<=N;i++)
	{
	    kv = kv*s + K[i];
	    QK[i] = kv;
	}

	if (Math.abs(kv)>Math.abs(K[N])*10.*ETA)
	{
	    // USE THE SCALED FORM OF THE RECURRENCE IF THE VALUE
	    // OF K AT S IS NONZERO
	    t = -pv/kv;
	    K[1] = QP[1];
	    for(int i=2;i<=N;i++)
	    {
		K[i] = t*QK[i-1] + QP[i];
	    }
	}
	else
	{
	    // USE UNSCALED FORM
	    K[1] = 0.0;
	    for(int i=2;i<=N;i++)
	    {
		K[i] = QK[i-1];
	    }
	}
	kv = K[1];
	for(int i=2;i<=N;i++)
	{
	    kv = kv*s + K[i];
	}
	t=Math.abs(kv)>Math.abs(K[N])*10.0*ETA?-pv/kv:0;
	s+=t;
    }
    return 0;
}


/**
 THIS ROUTINE CALCULATES SCALAR QUANTITIES USED TO
 COMPUTE THE NEXT K POLYNOMIAL AND NEW ESTIMATES OF
 THE QUADRATIC COEFFICIENTS.
 @return  TYPE - INTEGER VARIABLE SET HERE INDICATING HOW THE
 CALCULATIONS ARE NORMALIZED TO AVOID OVERFLOW.
*/
private int calcsc(final int N)
{
    // SYNTHETIC DIVISION OF K BY THE QUADRATIC 1,U,V
    quadsd(N, this.U, this.V, K, QK);
    this.C=quadsd_A;
    this.D=quadsd_B;
    if(_TRACE>0){
	System.err.println("calcsc: C="+C+"D="+D+
			   " K[N]="+K[N]+
			   " K[N-1]="+K[N-1]);
    }
    if (!(Math.abs(this.C)>Math.abs(K[N])*100.0*ETA) &&
	!(Math.abs(this.D)>Math.abs(K[N-1])*100.0*ETA))
    {
	// TYPE=3 INDICATES THE QUADRATIC IS ALMOST A FACTOR OF K
	return 3;
    }
    else if (!(Math.abs(this.D)<Math.abs(this.C)))
    {
	// TYPE=2 INDICATES THAT ALL FORMULAS ARE DIVIDED BY D
	this.E = this.A/this.D;
	this.F = this.C/this.D;
	this.G = this.U*this.B;
	this.H = this.V*this.B;
	this.A3 = (this.A+this.G)*this.E + this.H*(this.B/this.D);
	this.A1 = this.B*this.F - this.A;
	this.A7 = (this.F+this.U)*this.A + this.H;
	return 2;
    }
    else
    {
	// TYPE=1 INDICATES THAT ALL FORMULAS ARE DIVIDED BY C
	this.E = this.A/this.C;
	this.F = this.D/this.C;
	this.G = this.U*this.E;
	this.H = this.V*this.B;
	this.A3 = this.A*this.E + (this.H/this.C+this.G)*this.B;
	this.A1 = this.B - this.A*(this.D/this.C);
	this.A7 = this.A + this.G*this.D + this.H*this.F;
	return 1;
    }
}


/** COMPUTES THE NEXT K POLYNOMIALS USING SCALARS
    COMPUTED IN CALCSC.
*/
private void nextk(final int N,final int type)
{
      if (type==3)
      {
	  // USE UNSCALED FORM OF THE RECURRENCE IF TYPE IS 3
	  K[1] = 0.0;
	  K[2] = 0.0;
	  for(int i=3;i<=N;i++)
	  {
	      K[i] = QK[i-2];
	  }
	  return;
      }
      final double temp=type==1?this.B:this.A;

      if (Math.abs(this.A1)>Math.abs(temp)*ETA*10.0)
      {
	  // USE SCALED FORM OF THE RECURRENCE
	  this.A7 = this.A7/this.A1;
	  this.A3 = this.A3/this.A1;
	  K[1] = QP[1];
	  K[2] = QP[2] - this.A7*QP[1];
	  for(int i=3;i<=N;i++)
	  {
	      K[i] = this.A3*QK[i-2] - this.A7*QP[i-1] + QP[i];
	  }
	  return;
      }

      // IF A1 IS NEARLY ZERO THEN USE A SPECIAL FORM OF THE RECURRENCE
      K[1] = 0.0;
      K[2] = -this.A7*QP[1];
      for(int i=3;i<=N;i++)
      {
	  K[i] = this.A3*QK[i-2] - this.A7*QP[i-1];
      }
      return;
}


private double newest_uu,newest_vv;
/**
 COMPUTE NEW ESTIMATES OF THE QUADRATIC COEFFICIENTS
 USING THE SCALARS COMPUTED IN CALCSC.
*/
private void newest(final int N,final int NN,final int type)
{

    // USE FORMULAS APPROPRIATE TO SETTING OF TYPE.
    if (type==3)
    {
	// IF TYPE=3 THE QUADRATIC IS ZEROED
	newest_uu = newest_vv = 0.0;
	return;
    }
    final double a4,a5;
    if (type==2)
    {
	a4 = (this.A+this.G)*this.F + this.H;
	a5 = (this.F+this.U)*this.C + this.V*this.D;
    }
    else
    {
	a4 = this.A + this.U*this.B + this.H*this.F;
	a5 = this.C + (this.U+this.V*this.F)*this.D;
    }

    // EVALUATE NEW QUADRATIC COEFFICIENTS.
    final double b1 = -K[N]/P[NN];
    final double b2 = -(K[N-1]+b1*P[N])/P[NN];
    final double c1 = this.V*b2*this.A1;
    final double c2 = b1*this.A7;
    final double c3 = b1*b1*this.A3;
    final double c4 = c1 - c2 - c3;
    final double temp = a5 + b1*a4 - c4;
    if (!(temp!=0.0))
    {
	newest_uu = newest_vv = 0.0;
	return;
    }
    newest_uu = this.U - (this.U*(c3+c2)+this.V*(b1*this.A1+b2*this.A7))/temp;
    newest_vv = this.V*(1.0+c4/temp);
}


/** DIVIDES P BY THE QUADRATIC  1,U,V  PLACING THE
    QUOTIENT IN Q AND THE REMAINDER IN quadsd_A,quadsd_B
*/
double quadsd_A=0,quadsd_B=0;
private void quadsd(final int NN, final double u, final double v,final double [] p, final double [] q)
{
    double b=q[1]=p[1],a=q[2]=p[2]-u*b;
    for(int i=3;i<=NN;i++)
    {
        final double c = p[i] - u*a - v*b;
        q[i] = c;
        b = a;
        a = c;
    }
    quadsd_A = a;
    quadsd_B = b;
}

/** CALCULATE THE ZEROS OF THE QUADRATIC A*Z**2+B1*Z+C.
    THE QUADRATIC FORMULA, MODIFIED TO AVOID
    OVERFLOW, IS USED TO FIND THE LARGER ZERO IF THE
    ZEROS ARE REAL AND BOTH ZEROS ARE COMPLEX.
    THE SMALLER REAL ZERO IS FOUND DIRECTLY FROM THE
    PRODUCT OF THE ZEROS C/A.
*/
private void quad(final double a,final double b1,final double c)
{
this.SZR=this.SZI=this.LZR=this.LZI=0;
if (a==0.0)
{
    this.SZR = b1!=0?-(c/b1):0;
    return;
}
if (c == 0)
{
    this.LZR = -(b1/a);
    return;
}

// COMPUTE DISCRIMINANT AVOIDING OVERFLOW
final double absc=Math.abs(c),b=b1/2.0,e,d;
if (Math.abs(b)<absc)
{
    e = b*(b/absc) - (c<0?-a:a);
    d = Math.sqrt(Math.abs(e))*Math.sqrt(absc);
}
else
{
    e = 1.0 - (a/b)*(c/b);
    d = Math.sqrt(Math.abs(e))*Math.abs(b);
}
if(e<0)
{
    // COMPLEX CONJUGATE ZEROS
    this.LZR = this.SZR= -(b/a);
    this.SZI = Math.abs(d/a);
    this.LZI = -(this.SZI);
}
else
{
    // REAL ZEROS
    this.LZR = (-b + (b>=0?-d:d))/a;
    this.SZR = !(this.LZR==0)? (c/this.LZR)/a:0;
}
if(_TRACE>0)
{
    System.err.println("quad: a="+a+" b1="+b1+" c="+c+
		       " SZR="+SZR+" SZI="+SZI+" LZR="+LZR+" LZI="+LZI);
}

}


/** Taken originally from http://www.netlib.org/toms/419
    and adopted to fix broken scaling.
    RETURNS A SCALE FACTOR TO MULTIPLY THE COEFFICIENTS OF THE
    POLYNOMIAL. THE SCALING IS DONE TO AVOID OVERFLOW AND TO AVOID
    UNDETECTED UNDERFLOW INTERFERING WITH THE CONVERGENCE
    CRITERION.  THE FACTOR IS A POWER OF THE BASE.
    PT - MODULUS OF COEFFICIENTS OF P
    ETA,INFIN,SMALNO,BASE - CONSTANTS DESCRIBING THE
    FLOATING POINT ARITHMETIC.
*/
private static double pscale(final int NN,final double [] P)
{
    // FIND LARGEST AND SMALLEST MODULI OF COEFFICIENTS.
    double max=0,min=INFIN;
    for(int i=1;i<=NN;i++)
    {
	final double x=Math.abs(P[i]);
	if(x>max) max=x;
	if(x!=0 && x<min) min=x;
    }
    if(_TRACE>0)
    {
	System.err.println("PSCALE: NN="+NN+" MIN="+min+" MAX="+max);
    }
    // SCALE ONLY IF THERE ARE VERY LARGE OR VERY SMALL COMPONENTS.
    if(min>=LO && max<=HI)
    {
	return 1;
    }
    final double x=LO/min,sc;
    if(!(x>1))
    {
	sc=1.0/(Math.sqrt(max)*Math.sqrt(min));
    }
    else
    {
	sc=INFIN/x>max?1:x;
    }
    return Math.pow(BASE,(int)(Math.log(sc)/LOG_BASE + 0.5));
}



private void dumpc(final String s)
{
    System.err.println("---- DUMP:"+s);
    System.err.println("SR="+SR);
    System.err.println("SI="+SI);
    System.err.println("U="+U);
    System.err.println("V="+V);
    System.err.println("A="+A);
    System.err.println("B="+B);
    System.err.println("C="+C);
    System.err.println("D="+D);

    System.err.println("A1="+A1);
    System.err.println("A2="+A2);
    System.err.println("A3="+A3);
    System.err.println("A7="+A7);
    System.err.println("E="+E);
    System.err.println("F="+F);
    System.err.println("G="+G);
    System.err.println("H="+G);
    System.err.println("SZR="+SZR);
    System.err.println("SZI="+SZI);
    System.err.println("LZR="+LZR);
    System.err.println("LZI="+LZI);
    System.err.println("---- DUMP END");
    System.err.println();
}



/** Unit Test. */
public static void main(String [] args)
{
    final long t0=System.currentTimeMillis();

    final double eps=1e-4;
    final int N_MAX=6;
    if(args.length>0)
    { // test from file
	final PolynomialRootJenkinsTraubORIG rootfinder=
		new PolynomialRootJenkinsTraubORIG(N_MAX);
	    final com.polytechnik.utils.PolynomialRootsTester t=
		new com.polytechnik.utils.PolynomialRootsTester(
		    rootfinder,eps);
	    for(String f:args)
	    {
		try {
		    t.testPolynomialFromFile(f);
		} catch(final java.io.IOException e){
		    throw new RuntimeException(e);
		}
	    }
	}
	else
	{
	    final java.util.Random rn=new java.util.Random(-1381923);
	    final int n_tests=100000,N_MIN=2;
	    final PolynomialRootJenkinsTraubORIG rootfinder=
		new PolynomialRootJenkinsTraubORIG(N_MAX);

	    for(int n=N_MIN;n<N_MAX;n++)
	    {
		final com.polytechnik.utils.PolynomialRootsTester t=
		    new com.polytechnik.utils.PolynomialRootsTester(
			rootfinder,eps);
		t.testRootsRandomPolynomial(n,n_tests,rn);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("PolynomialRootsCompanionMatrix(): tests OK done in "+
			   (t1-t0)+" milliseconds. ver=$Id: PolynomialRootJenkinsTraubORIG.java,v 1.66 2011/11/06 07:42:46 mal Exp $");
    }

}


