package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Polynomial root in arbitrary basis.
 *  @version $Id: PolynomialRootsConfederateMatrixABasis.java,v 1.21 2015/10/23 19:25:31 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class PolynomialRootsConfederateMatrixABasis
           implements PolynomialRootsFinder {

    /** Get confederate matrix for the basis we use.
     * For monomials basis - this is usual companion matrix,
     * for other basises, e.g. orthogonal polynomials, see:
     * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
     * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
     * and resulting recurrence relations and characterizations", March 2007.
     * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
     */
    public abstract double [] getConfederateMatrix(final double [] polynomial_coefs);
    public int findRoots(final int n,
			 final double [] polynomial_coefs,
			 final double [] re_root,
			 final double [] im_root)
    {
	final double a=polynomial_coefs[n];
	if(!(a!=0))
	{
	    java.util.Arrays.fill(re_root,0,n,Double.NaN);
	    java.util.Arrays.fill(im_root,0,n,Double.NaN);
	    return -1;
	}
	final double [] coefs;
	if(n+1!=polynomial_coefs.length){
	    //throw new RuntimeException("shrink array, n="+n+" polynomial_coefs.length="+polynomial_coefs.length);
	    coefs=new double [n+1];
	    System.arraycopy(polynomial_coefs,0,coefs,0,n+1);
	}
	else coefs=polynomial_coefs;
	final double [] matr=getConfederateMatrix(coefs);

	// note, that fortran notation is a transponse (first index change most often)
	// to java one, which we use (last index change most often),
	// but eigenvalues are the same and we do not need eigenvectors here.
	final int info=com.polytechnik.lapack.Eigenvalues.getNonSymEV_dgees(n,matr,n,re_root,im_root);

	if(info!=0)
	{
	    java.util.Arrays.fill(re_root,0,n,Double.NaN);
	    java.util.Arrays.fill(im_root,0,n,Double.NaN);
	    return -1;
	}
	return n;
    }

static class PolynomialRootsConfederateMatrixPBasis extends PolynomialRootsConfederateMatrixABasis {
    final BasisPolynomials B;
    PolynomialRootsConfederateMatrixPBasis(final BasisPolynomials B){this.B=B;}
    public double [] getConfederateMatrix(final double [] polynomial_coefs)
    {return B.getConfederateMatrix(polynomial_coefs);}
}

public static final PolynomialRootsFinder
ABASISMONOMIALS=new PolynomialRootsConfederateMatrixPBasis(new Monomials()),
ABASISLAGUERRE=new PolynomialRootsConfederateMatrixPBasis(new Laguerre()),
ABASISLEGENDRESHIFTED=new PolynomialRootsConfederateMatrixPBasis(new LegendreShifted()),
ABASISCHEBYSHEV=new PolynomialRootsConfederateMatrixPBasis(new Chebyshev()),
ABASISLEGENDRE=new PolynomialRootsConfederateMatrixPBasis(new Legendre()),
ABASISHERMITEE=new PolynomialRootsConfederateMatrixPBasis(new HermiteE());


private static void test_Solvers(final int n,final int n_tests,final java.util.Random r,final double eps)
{
    for(int kk=0;kk<n_tests;kk++)
    {

    final double [] pol=new double [n+1];
    for(int k=pol.length;--k>=0;) pol[k]=(1-2*r.nextDouble())/(k+1);
    pol[n]+=3.2;

    int nroots;
    final double [] root_re=new double [n],root_im=new double [n];
    nroots=new PolynomialRootJenkinsTraubORIG(n).findRoots(n,pol,root_re,root_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root_re,root_im),root_re,root_im);


    final double [] root1_re=new double [n],root1_im=new double [n];
    nroots=ABASISMONOMIALS.findRoots(n,pol,root1_re,root1_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root1_re,root1_im),root1_re,root1_im);
    UnitTests.checkDoubleArrayRel1(root1_re,"root1_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root1_im,"root1_im",root_im,"root_im",eps);


    final double [] root2_re=new double [n],root2_im=new double [n];
    nroots=ABASISLAGUERRE.findRoots(n,new Laguerre().convertToBasisFromPolynomial_Test(n,pol),root2_re,root2_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root2_re,root2_im),root2_re,root2_im);
    UnitTests.checkDoubleArrayRel1(root2_re,"root2_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root2_im,"root2_im",root_im,"root_im",eps);


    final double [] root3_re=new double [n],root3_im=new double [n];
    nroots=ABASISLEGENDRESHIFTED.findRoots(n,new LegendreShifted().convertToBasisFromPolynomial_Test(n,pol),root3_re,root3_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root3_re,root3_im),root3_re,root3_im);
    UnitTests.checkDoubleArrayRel1(root3_re,"root3_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root3_im,"root3_im",root_im,"root_im",eps);


    final double [] root4_re=new double [n],root4_im=new double [n];
    nroots=ABASISCHEBYSHEV.findRoots(n,new Chebyshev().convertToBasisFromPolynomial_Test(n,pol),root4_re,root4_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root4_re,root4_im),root4_re,root4_im);
    UnitTests.checkDoubleArrayRel1(root4_re,"root4_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root4_im,"root4_im",root_im,"root_im",eps);

    final double [] root5_re=new double [n],root5_im=new double [n];
    nroots=ABASISLEGENDRE.findRoots(n,new Legendre().convertToBasisFromPolynomial_Test(n,pol),root5_re,root5_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root5_re,root5_im),root5_re,root5_im);
    UnitTests.checkDoubleArrayRel1(root5_re,"root5_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root5_im,"root5_im",root_im,"root_im",eps);


    final double [] root6_re=new double [n],root6_im=new double [n];
    nroots=ABASISHERMITEE.findRoots(n,new HermiteE().convertToBasisFromPolynomial_Test(n,pol),root6_re,root6_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root6_re,root6_im),root6_re,root6_im);
    UnitTests.checkDoubleArrayRel1(root6_re,"root6_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root6_im,"root6_im",root_im,"root_im",eps);


    final double [] root7_re=new double [n],root7_im=new double [n];
    nroots=ABASISMONOMIALS.findRoots(n,new Monomials().convertToBasisFromPolynomial_Test(n,pol),root7_re,root7_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWith2Load(nroots,TMatr.sumAB(nroots,root7_re,root7_im),root7_re,root7_im);
    UnitTests.checkDoubleArrayRel1(root7_re,"root7_re",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(root7_im,"root7_im",root_im,"root_im",eps);

    }

    System.err.println("PolynomialRootsConfederateMatrixABasis:test_Solvers() test OK for N="+n+" n_tests="+n_tests+" with eps="+eps);

}

private static void printAllQuadratures(final int n_max,final java.util.Random r)
{
    for(int n=1;n<=n_max;n++)
    {
	if(true)
	{
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsABasis.printQuadratures("LegendreShifted",n,moments,new OrthogonalPolynomialsLegendreShiftedBasis(),false);
	}
	if(true)
	{
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsABasis.printQuadratures("Laguerre",n,moments,new OrthogonalPolynomialsLaguerreBasis(),false);
	}
	if(true)
	{
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsABasis.printQuadratures("Legendre",n,moments,new OrthogonalPolynomialsLegendreBasis(),false);
	}
	if(true)
	{
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsABasis.printQuadratures("Chebyshev",n,moments,new OrthogonalPolynomialsChebyshevBasis(),false);
	}
	if(true)
	{
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsABasis.printQuadratures("HermiteE",n,moments,new OrthogonalPolynomialsHermiteEBasis(),false);
	}
    }
}


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(8631);

	printAllQuadratures(80,r);

	final double eps=1e-4;
	for(int n=1;n<=8;n++)
	{
	    test_Solvers(n,110000,r,eps);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("PolynomialRootsConfederateMatrix(): tests OK done in "+
			   (t1-t0)+" milliseconds. ver=$Id: PolynomialRootsConfederateMatrixABasis.java,v 1.21 2015/10/23 19:25:31 mal Exp $");
    }
}
