package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** 3 points stable Cholesky elimination.
 *  Algorithm from:
 *  A.A.Samarskiy, E.S.Nikolaev, Grid Calculations Methods, 1978, p. 73.
 *  @version $Id: Progonka3.java,v 1.22 2012/04/19 21:47:01 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Progonka3  {


    /** 3 non-zero points in a row.
     *  @param am diag-1 element.
     *  @param cm diag element.
     *  @param bm diag+1 element.
     *  @param fm right part.
     *  @param y solution.
     */
    public static void eliminate(final int n,
				 final double [] am,
				 final double [] cm,
				 final double [] bm,
				 final double [] fm,
				 final double [] y,

				 final double [] alfa,
				 final int [] in1,
				 final int [] in2)
    {
	if(n==0) return;
	else if(n==1) {y[0]=fm[0]/cm[0];return;}
	if(n<0){throw new IllegalArgumentException("n="+n);}
	/*
	// for debugging
	java.util.Arrays.fill(in1,-1);
	java.util.Arrays.fill(in2,-1);
	java.util.Arrays.fill(y,Double.NaN);
	java.util.Arrays.fill(alfa,Double.NaN);
	*/


	final int n1=n-1;
	double c=cm[0],f=fm[0],a=am[1],u=fm[1];
	int prev_in2=0;
	for(int i=0;i<n1;i++)
	{
	    final int i1=i+1,i2=i1+1;
	    final double bm_i=bm[i],cm_i1=cm[i1];
	    if(Math.abs(c)>=Math.abs(bm_i))
	    {
		final double ali=alfa[i]=bm_i/c,bei=y[prev_in2]=f/c;
		in1[i]=prev_in2;
		prev_in2=in2[i]=i1;

		c=cm_i1-a*ali;
		f=u-a*bei;
		if(i2>=n)
		{
		    continue;
		}
		a=am[i2];
		u=fm[i2];
	    }
	    else
	    {
		final double ali=alfa[i]=c/bm_i,bei=y[i1]=f/bm_i;
		in1[i]=i1;
		in2[i]=prev_in2;

		c=a-cm_i1*ali;
		f=u-cm_i1*bei;
		if(i2>=n)
		{
		    continue;
		}
		a=-ali*am[i2];
		u=fm[i2]-am[i2]*bei;
	    }
	}

	y[prev_in2]=f/c;
	for(int i=n1;i-->0;)
	{
	    // alfas are saved with off by 1 index to allow re-use of bm as temporary array alfa.
	    y[in1[i]]-=alfa[i]*y[in2[i]];
	}
    }



    /**
     * Test.
     * java   com/polytechnik/utils/Progonka3
     * checkLinSystem3: 200000 tests OK done in 80353 milliseconds. ver=$Id: Progonka3.java,v 1.22 2012/04/19 21:47:01 mal Exp $
     */
    private static void checkLinSystem3()
    {
	// a test with random matrix
	final java.util.Random r=new java.util.Random(-179093);
	final long t0=System.currentTimeMillis();
	final double eps=1e-5;
	int i=0;
	for(;i<200000;i++)
	{
	    final int n=(i<1000?r.nextInt(5):r.nextInt(1+5000));
	    final double [] a=new double [n],b=new double [n],c=new double [n],
		f=new double [n],y=new double [n],al=new double [n];
	    final int [] i1=new int [n],i2=new int [n];
	    for(int j=0;j<n;j++)
	    {
		a[j]=r.nextDouble();
		c[j]=r.nextDouble();
		b[j]=r.nextDouble();
		f[j]=r.nextDouble();
	    }

	    eliminate(n,a,c,b,f,y,al,i1,i2);
	    double s=0;
	    if(n>0) s+=Math.abs(c[0]*y[0]+(n>1?b[0]*y[1]:0)-f[0]);
	    if(n>1) s+=Math.abs(a[n-1]*y[n-2]+c[n-1]*y[n-1]-f[n-1]);
	    for(int j=1;j<n-1;j++)
	    {
		s+=Math.abs(a[j]*y[j-1]+c[j]*y[j]+b[j]*y[j+1]-f[j]);
	    }
	    if(s>eps||Double.isNaN(s))
	    {
		throw new RuntimeException("s="+s+" n="+n);
	    }


	    // try with storage re-usage
	    final double [] bs=b.clone(),ys=f.clone();
	    eliminate(n,a,c,bs,ys,ys,bs,i1,i2);
	    for(int j=0;j<n;j++)
	    {
		if(!(ys[j]==y[j]))
		{
		    throw new RuntimeException("ys["+j+"="+ys[j]+" y["+j+"="+y[j]+" n="+n);
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("checkLinSystem3: "+i+" tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: Progonka3.java,v 1.22 2012/04/19 21:47:01 mal Exp $");
    }




    public static void main(String [] args)
    {
	checkLinSystem3();
    }
}
