package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** NxM points Vandermonde function interpolation.
 *  @version $Id: Vandermonde2Interpolation.java,v 1.11 2011/11/05 19:31:52 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class Vandermonde2Interpolation  {
    final int order_X,order_Y;
    private final double [][] matrix;
    private final double [] rightpart;
    private final int [] ipvt;
    public Vandermonde2Interpolation(final int order_X,final int order_Y)
    {
	this.order_X=order_X;
	this.order_Y=order_Y;
	final int n=this.order_X*this.order_Y;
	this.rightpart=new double [n];
	this.matrix=getMatrix(this.order_X,this.order_Y);
	// prepare matrix for linear equation solving
	this.ipvt=new int [n];
	final int info=Linsystems.dgefa(this.matrix,n,this.ipvt);
	if(info!=0)
	{
	    throw new RuntimeException("INFO="+info);
	}
    }
    protected double [][] getMatrix(final int o_X,final int o_Y)
    {
	//System.err.println(" === o_X="+o_X+" o_Y="+o_Y);
	final int n=o_X*o_Y;
	final double [][] m=new double [n][];
	for(int i=0,c=0;i<o_X;i++)
	{
	    final double x=((double)(i+1))/(1+o_X); // equal distance
	    for(int j=0;j<o_Y;j++,c++)
	    {
		final double [] l=m[c]=new double [n];
		final double y=((double)(j+1))/(1+o_Y); // equal distance
		double mx=1;
		for(int i1=0,c1=0;i1<o_X;i1++)
		{
		    double my=mx;
		    for(int j1=0;j1<o_Y;j1++,c1++)
		    {
			l[c1]=my;
			//System.err.println("i1="+i1+" j1="+j1+" l[c1]="+l[c1]);
			my*=y;
		    }
		    mx*=x;
		}
	    }
	}
	return m;
    }

    public void setMoments(final double [][] moments)
    {
	if(moments.length!=this.order_X)
	{
	    throw new RuntimeException("moments.length="+moments.length+
				       " should be this.order_X="+(this.order_X));
	}
	for(int i=0,off=0;i<this.order_X;i++,off+=this.order_Y)
	{
	    if(moments[i].length!=this.order_Y)
	    {
		throw new RuntimeException("moments["+i+"].length="+moments[i].length+
					   " should be this.order_Y="+(this.order_Y));
	    }
	    System.arraycopy(moments[i],0,this.rightpart,off,this.order_Y);
	}
	Linsystems.dgesl(this.matrix,this.rightpart.length,this.ipvt,this.rightpart,1);
    }

    public double getF(final double x,final double y)
    {
	double sx=0;
	for(int i=this.order_X,r=this.rightpart.length-1;i-->0;)
	{
	    double sy=0;
	    for(int j=this.order_Y;j-->0;r--)
	    {
		sy=y*sy+this.rightpart[r];
	    }
	    sx=x*sx+sy;
	}
	return sx;
    }
    public double getDFxFy(final double x,final double y)
    {
	double sx=0;
	for(int i=this.order_X;i-->1;)
	{
	    int r=(i+1)*this.order_Y-1;
	    double sy=0;
	    for(int j=this.order_Y;j-->0;r--)
	    {
		sy=y*sy+this.rightpart[r];
	    }
	    sx=x*sx+sy*i;
	}
	return sx;
    }
    public double getFxDFy(final double x,final double y)
    {
	double sx=0;
	for(int i=this.order_X;i-->0;)
	{
	    int r=(i+1)*this.order_Y-1;
	    double sy=0;
	    for(int j=this.order_Y;j-->1;r--)
	    {
		sy=y*sy+this.rightpart[r]*j;
	    }
	    sx=x*sx+sy;
	}
	return sx;
    }
    public double getDFxDFy(final double x,final double y)
    {
	double sx=0;
	for(int i=this.order_X;i-->1;)
	{
	    int r=(i+1)*this.order_Y-1;
	    double sy=0;
	    for(int j=this.order_Y;j-->1;r--)
	    {
		sy=y*sy+this.rightpart[r]*j;
	    }
	    sx=x*sx+sy*i;
	}
	return sx;
    }
    public double getD2FxD2Fy(final double x,final double y)
    {
	double sx=0;
	for(int i=this.order_X;i-->2;)
	{
	    int r=(i+1)*this.order_Y-1;
	    double sy=0;
	    for(int j=this.order_Y;j-->2;r--)
	    {
		sy=y*sy+this.rightpart[r]*(j*(j-1));
	    }
	    sx=x*sx+sy*(i*(i-1));
	}
	return sx;
    }



    static void checkValues(final Vandermonde2Interpolation v)
    {
	final double [][] mom=new double [v.order_X][v.order_Y];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-5;
	int n=0;
	for(;n<100000;n++)
	{
	    for(int i=0;i<mom.length;i++)
	    {
		final double [] mo=mom[i];
		for(int j=0;j<mo.length;j++)
		{
		    mo[j]=r.nextDouble();
		}
	    }
	    v.setMoments(mom);


	    for(int i=0;i<mom.length;i++)
	    {
		final double [] mo=mom[i];
		for(int j=0;j<mo.length;j++)
		{
		    final double fi=v.getF(((double)(i+1))/(1+v.order_X),((double)(j+1))/(1+v.order_Y));
		    UnitTests.checkDouble(fi,"fi",mom[i][j],"mom[i][j]",eps);
		}
	    }
	}
	System.err.println("checkValues() "+n+" test OK values matched"+
			   " order_X="+v.order_X+
			   " order_Y="+v.order_Y+".");
    }

    private static void checkExactPolinomial(final Vandermonde2Interpolation v)
    {
	final double [][] mom=new double [v.order_X][v.order_Y],coefs=new double [v.order_X][v.order_Y];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-5;

	int n=0;
	for(;n<1000;n++)
	{
	    // random polinomian
	    for(int i=0;i<coefs.length;i++)
	    {
		for(int j=0;j<coefs[i].length;j++)
		{
		    coefs[i][j]=r.nextDouble();
		}
	    }
	    for(int i=0;i<coefs.length;i++)
	    {
		final double x=((double)(i+1))/(1+v.order_X);
		for(int j=0;j<coefs[i].length;j++)
		{
		    final double y=((double)(j+1))/(1+v.order_Y);
		    double s=0;
		    for(int i1=0;i1<coefs.length;i1++)
		    {
			for(int j1=0;j1<coefs[i1].length;j1++)
			{
			    s+=coefs[i1][j1]*Math.pow(x,i1)*Math.pow(y,j1);
			}
		    }
		    mom[i][j]=s;
		}
	    }
	    v.setMoments(mom);

	    // match exact values of v.order-1 polynomial
	    for(double x=-0.5;x<1;x+=5e-2)
	    {
		for(double y=-0.5;y<1;y+=5e-2)
		{
		    final double f=v.getF(x,y),
			dfxfy=v.getDFxFy(x,y),
			fxdfy=v.getFxDFy(x,y),
			dfxdfy=v.getDFxDFy(x,y),
			d2fxd2fy=v.getD2FxD2Fy(x,y);

		    double sf=0,sdfxfy=0,sfxdfy=0,sdfxdfy=0,sd2fxd2fy=0;
		    for(int i1=0;i1<coefs.length;i1++)
		    {
			for(int j1=0;j1<coefs[i1].length;j1++)
			{
			    sf+=coefs[i1][j1]*Math.pow(x,i1)*Math.pow(y,j1);
			    if(i1>0)
			    {
				sdfxfy+=i1*coefs[i1][j1]*Math.pow(x,i1-1)*Math.pow(y,j1);
			    }
			    if(j1>0)
			    {
				sfxdfy+=j1*coefs[i1][j1]*Math.pow(x,i1)*Math.pow(y,j1-1);
			    }
			    if(j1>0&&i1>0)
			    {
				sdfxdfy+=i1*j1*coefs[i1][j1]*Math.pow(x,i1-1)*Math.pow(y,j1-1);
			    }
			    if(j1>1&&i1>1)
			    {
				sd2fxd2fy+=(i1*(i1-1)*j1*(j1-1))*coefs[i1][j1]*Math.pow(x,i1-2)*Math.pow(y,j1-2);
			    }
			}
		    }

		    // match values
		    UnitTests.checkDouble(sf,"sf",f,"f",eps);
		    UnitTests.checkDouble(sdfxfy,"sdfxfy",dfxfy,"dfxfy",eps);
		    UnitTests.checkDouble(sfxdfy,"sfxdfy",fxdfy,"fxdfy",eps);
		    UnitTests.checkDouble(sdfxdfy,"sdfxdfy",dfxdfy,"dfxdfy",eps);
		    UnitTests.checkDouble(sd2fxd2fy,"sd2fxd2fy",d2fxd2fy,"d2fxd2fy",eps);
		}
	    }
	}
	System.err.println("checkExactPolinomial() "+n+" test OK values matched"+
			   " order_X="+v.order_X+
			   " order_Y="+v.order_Y+".");
    }


    private static final String S_F="F",S_DFXFY="dfxfy",S_FXDFY="fxdfy",S_DFXDFY="dfxdfy",S_D2FXD2FY="d2fxd2fy";

    static void printCoefs(final String what,
			   final Vandermonde2Interpolation v,
			   final double x,
			   final double y)
    {
	System.err.println("printCoefs coefficients "+what+" for"+
			   " order_X="+v.order_X+
			   " order_Y="+v.order_Y+
			   " x="+x+" y="+y);
	final double hx=1.0/(1+v.order_X),hy=1.0/(1+v.order_Y);
	for(int i=0;i<v.order_X;i++)
	{
	    System.err.print(String.valueOf(i)+":");

	    for(int j=0;j<v.order_Y;j++)
	    {
		final double [][] mom=new double [v.order_X][v.order_Y];
		mom[i][j]=1;
		v.setMoments(mom);
		System.err.print("\t");
		if(what==S_F)
		{
		    System.err.print(v.getF(x,y));
		}
		else if(what==S_FXDFY)
		{
		    System.err.print(v.getFxDFy(x,y)*hy);
		}
		else if(what==S_DFXFY)
		{
		    System.err.print(v.getDFxFy(x,y)*hx);
		}
		else if(what==S_DFXDFY)
		{
		    System.err.print(v.getDFxDFy(x,y)*(hx*hy));
		}
		else if(what==S_D2FXD2FY)
		{
		    System.err.print(v.getD2FxD2Fy(x,y)*(hx*hy*hx*hy));
		}
		else
		{
		    throw new IllegalArgumentException();
		}
	    }
	    System.err.println();
	}
    }


    public static void main(String [] s)
    {
	checkValues(new Vandermonde2Interpolation(1,5));
	checkValues(new Vandermonde2Interpolation(5,1));
	checkValues(new Vandermonde2Interpolation(2,2));
	checkValues(new Vandermonde2Interpolation(3,4));
	checkValues(new Vandermonde2Interpolation(7,6));
	checkValues(new Vandermonde2Interpolation(5,3));
	checkExactPolinomial(new Vandermonde2Interpolation(3,5));

	printCoefs(S_DFXDFY,new Vandermonde2Interpolation(2,2),0.5,0.5);
	printCoefs(S_DFXDFY,new Vandermonde2Interpolation(3,3),0.5,0.5);
	printCoefs(S_DFXDFY,new Vandermonde2Interpolation(3,2),0.5,0.5);
	printCoefs(S_D2FXD2FY,new Vandermonde2Interpolation(3,3),0.5,0.5);
	printCoefs(S_FXDFY,new Vandermonde2Interpolation(1,5),0.5,1-1.0/(1+5));
	printCoefs(S_FXDFY,new Vandermonde2Interpolation(5,5),1-1.0/(1+5),1-1.0/(1+5));
	printCoefs(S_FXDFY,new Vandermonde2Interpolation(5,5),1,1);
    }
}
