package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;

/** Volatility calculation.
 *  Unit test is in VkQVM, BkQVM and XkVM.XkQVMTest.test_Matrices()
 *  @version $Id: VolMatrix.java,v 1.4 2015/09/06 08:46:40 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class VolMatrix {

/** Calculated D(psi).
 */
public abstract double [] getDPsi(final double [] psi);

/** Volatility matrix. D^T QQ D
 */
public double [] getVolMatrix(final int n,final double [] QQ,final int ldQQ)
{
    final double [] DQQD=new double [n*n];
    final double [][] bb=new double [n][];
    for(int k=0;k<n;k++)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=bb[k]=getDPsi(pp);
	if(dpsik.length>pp.length) throw new RuntimeException("invalid getDPsi");
	TMatr.setMA_x_MB(pp,pp.length,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	for(int l=0;l<=k;l++)
	{
	    DQQD[n*k+l]=DQQD[n*l+k]=TMatr.mulVV(bb[l],pp);
	}
    }
    return DQQD;
}

/** Differentiated matrix. QQ D + D^T QQ^T.
 */
public double [] getDiffMatrix(final int n,final double [] QQ,final int ldQQ)
{
    final double [] QQD=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=getDPsi(pp);
	TMatr.setMA_x_MB(QQD,n,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	if(k>0) System.arraycopy(QQD,0,QQD,k*n,n);
    }
    TMatr.selectHermitianPart(n,QQD,n,0,false); // factor of 2 is also here
    return QQD;
}

/** D2 Differentiated matrix. QQ DD + DD^T QQ^T + D^T QQ D + D^T QQ^T D
 */
public double [] getDiff2Matrix(final int n,final double [] QQ,final int ldQQ)
{
    final double [] QQD2=new double [n*n],pps=new double [n];
    final double [][] bb=new double [n][];
    for(int k=n;--k>=0;){final double [] ppk=new double [k+1];ppk[k]=1;bb[k]=getDPsi(ppk);}
    for(int k=n;--k>=0;)
    {
	final double [] dpsik=bb[k],ddpsik=getDPsi(dpsik);
	TMatr.setMA_x_MB(QQD2,n,QQ,ldQQ,0,ddpsik.length,ddpsik,1,1,0);
	TMatr.setMA_x_MB(pps,n,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	for(int l=n,kn=k*n;--l>=0;){QQD2[kn+l]=QQD2[l]+TMatr.mulVV(bb[l],pps);}
    }
    TMatr.selectHermitianPart(n,QQD2,n,0,false); // factor of 2 is also here
    return QQD2;
}

/** A part of D2: QQ DD + DD^T QQ^T.
 */
public double [] getDiff2PartMatrix(final int n,final double [] QQ,final int ldQQ)
{
    final double [] QQPD2=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] ddpsik=getDPsi(getDPsi(pp));
	TMatr.setMA_x_MB(QQPD2,n,QQ,ldQQ,0,ddpsik.length,ddpsik,1,1,0);
	if(k>0) System.arraycopy(QQPD2,0,QQPD2,k*n,n);
    }
    TMatr.selectHermitianPart(n,QQPD2,n,0,false); // factor of 2 is also here
    return QQPD2;
}

/** D- matrix. If Dpsi=getDPsi(psi), then Dpsi_k=\sum_{l}D_{kl}psi_l.
 *
 */
public double [] getDMatrix(final int n)
{
    final double [] D=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=getDPsi(pp);
	for(int l=dpsik.length;--l>=0;) D[n*l+k]=dpsik[l];
    }
    return D;
}

/** Hermitian QQ D + D^T QQ^T.
 */
public static double [] getHAB(final int n,final double [] QQ,final double [] D)
{
    final double [] HAB=TMatr.getMatr_x_Matr(n,QQ,n,D,n);
    TMatr.selectHermitianPart(n,HAB,n,0,false); // factor of 2 is also here
    return HAB;
}

/** Calculates <r*Q_i Q_j>.
 */
public static double [] getRadauMatrix(final int n,final double [] QQ,final int ldQQ,final com.polytechnik.utils.OrthogonalPolynomialsABasis M)
{
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp);
	TMatr.setMA_x_MB(rQQ,nd,QQ,ldQQ,0,rpp.length,rpp,1,1,0);
	if(k>0) System.arraycopy(rQQ,0,rQQ,k*nd,nd);
    }
    return rQQ;
}

/** Calculates <D(r*Q_i Q_j)>.
 */
public double [] getDRadauMatrix(final int n,final double [] QQ,final int ldQQ,final com.polytechnik.utils.OrthogonalPolynomialsABasis M)
{
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd],tmp1=new double [nd],tmp2=new double [nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] Drpp=getDPsi(M.getRadauMultiplied(pp)),rDpp=M.getRadauMultiplied(getDPsi(pp));
	TMatr.setMA_x_MB(tmp1,nd,QQ,ldQQ,0,Drpp.length,Drpp,1,1,0);
	TMatr.setMA_x_MB(tmp2,1,rDpp,n,0,rDpp.length,QQ,nd,ldQQ,0);
	for(int j=nd;--j>=0;) {rQQ[k*nd+j]+=tmp1[j];rQQ[j*nd+k]+=tmp2[j];}
    }
    return rQQ;
}

/** Calculates <r^2*Q_i Q_j>.
 */
public static double [] getRadau2Matrix(final int n,final double [] QQ,final int ldQQ,final com.polytechnik.utils.OrthogonalPolynomialsABasis M)
{
    final int nd=n-1;
    final double [] trans=new double [n*nd],rQQ=new double [nd*nd],tmp=new double [n*nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp);
	if(rpp.length!=pp.length+1) throw new IllegalArgumentException();
	System.arraycopy(rpp,0,trans,k*n,rpp.length);
    }
    // calculate trans QQ trans^T, trans: nd x n.
    TMatr.setMA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,trans,nd,n,0); // tmp: n x nd
    TMatr.setMA_x_MB(rQQ,nd,trans,n,0,n,tmp,nd,nd,0);
    return rQQ;
}

/** Calculates <D(r^2*Q_i Q_j)>.
 */
public double [] getDRadau2Matrix(final int n,final double [] QQ,final int ldQQ,final com.polytechnik.utils.OrthogonalPolynomialsABasis M)
{
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd],transR=new double [nd*n],transD=new double [nd*n],tmp=new double [nd*n];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp),Drpp=getDPsi(rpp);
	if(rpp.length!=pp.length+1||Drpp.length>n) throw new IllegalArgumentException();
	System.arraycopy(rpp,0,transR,k*n,rpp.length);
	System.arraycopy(Drpp,0,transD,k*n,Drpp.length);
    }
    // transR: nd x n ; transD: nd x n
    // now calculate transR QQ transD^T + transD QQ transR^T
    TMatr.setMA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,transD,nd,n,0); // tmp: n x nd
    TMatr.setMA_x_MB(rQQ,nd,transR,n,0,n,tmp,nd,nd,0);
    // QQ is assumed to be symmetric
    if(true) TMatr.selectHermitianPart(nd,rQQ,nd,0,false);
    else{
	TMatr.setMA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,transR,nd,n,0); // tmp: n x nd
	TMatr.setMA_x_MB(transR,nd,transD,n,0,n,tmp,nd,nd,0);
	TMatr.sumP1(rQQ,transR,nd*nd,1);
    }
    return rQQ;
}

}
